import sys

if sys.platform.startswith("linux"):
    from opencanary.modules import CanaryService, FileSystemWatcher
    import re

    class SambaLogWatcher(FileSystemWatcher):
        def __init__(self, logFile=None, logger=None):
            self.logger = logger
            FileSystemWatcher.__init__(self, fileName=logFile)

        def handleLines(self, lines=None):
            audit_re = re.compile(r"^.*smbd_audit.*: (.*$)")

            for line in lines:
                matches = audit_re.match(line)

                # Skip lines that do not match the correct RegEx pattern
                if matches is None:
                    continue

                data = matches.groups()[0].split("|")

                user = data[0]
                srcHost = data[1]
                dstHost = data[2]
                srcHostName = data[3]
                shareName = data[4]
                dstHostName = data[5]
                smbVersion = data[6]
                smbArch = data[7]
                domainName = data[9]
                auditAction = data[10]
                auditStatus = data[11]
                path = data[12]

                if user == "":
                    user = "anonymous"

                data = {}
                data["src_host"] = srcHost
                data["src_port"] = "-1"
                data["dst_host"] = dstHost
                data["dst_port"] = 445
                data["logtype"] = self.logger.LOG_SMB_FILE_OPEN
                data["logdata"] = {
                    "USER": user,
                    "REMOTENAME": srcHostName,
                    "SHARENAME": shareName,
                    "LOCALNAME": dstHostName,
                    "SMBVER": smbVersion,
                    "SMBARCH": smbArch,
                    "DOMAIN": domainName,
                    "AUDITACTION": auditAction,
                    "STATUS": auditStatus,
                    "FILENAME": path,
                }
                self.logger.log(data)

    class CanarySamba(CanaryService):
        NAME = "smb"

        def __init__(self, config=None, logger=None):
            CanaryService.__init__(self, config=config, logger=logger)
            self.audit_file = config.getVal(
                "smb.auditfile", default="/var/log/samba-audit.log"
            )
            self.config = config

        def startYourEngines(self, reactor=None):
            # create samba run dir, so testparm doesn't error
            # try:
            #    os.stat('/var/run/samba')
            # except OSError:
            #    os.mkdir('/var/run/samba')

            fs = SambaLogWatcher(logFile=self.audit_file, logger=self.logger)
            fs.start()



import os
import datetime

from opencanary.modules import CanaryService

from base64 import b64decode

from twisted.application import internet
from twisted.web.http import Request, HTTPChannel
from twisted.web import http

from jinja2 import Template

PROFILES = {
    "ms-isa": {
        # p/Microsoft ISA Server Web Proxy/
        # Force HTTP/1.1 reply even when sent HTTP/1.0 (to match the nmap version sig)
        "HTTP1.1_always": True,
        "headers": [
            ("Via", "1.1 localhost"),
            ("Proxy-Authenticate", "Basic"),
            # TODO: more realistict authentication for ISA
            # ("Proxy-Authenticate", "NTLM"),
            # ("Proxy-Authenticate", "Kerberos"),
            # ("Proxy-Authenticate", "Negotiate"),
            ("Pragma", "no-cache"),
            ("Cache-Control", "no-cache"),
        ],
        "status_reason": b"Proxy Authentication Required ( The ISA Server requires authorization to fulfill the request. Access to the Web Proxy service is denied.  )",
    },
    "squid": {
        # p/Squid http proxy/ v/$1/ cpe:/a:squid-cache:squid:$1/
        "banner": "Squid proxy-caching web server",
        "headers": [
            ("Server", "squid/3.3.8"),
            ("Mime-Version", "1.0"),
            ("Vary", "Accept-Language"),
            ("Via", "1.1 localhost (squid/3.3.8)"),
            ("X-Cache", "MISS from localhost"),
            ("X-Cache-Lookup", "NONE from localhost"),  # actually hostname:port
            ("X-Squid-Error", "ERR_CACHE_ACCESS_DENIED 0"),
        ],
        "status_reason": b"Proxy Authentication Required",
    },
}


class AlertProxyRequest(Request):
    """
    Used by Proxy to implement a simple web proxy.
    """

    FACTORY = None

    def __init__(self, channel, queued):
        Request.__init__(self, channel, queued)

    def logAuth(self):
        auth = self.getHeader("Proxy-Authorization")
        if auth is None:
            return

        factory = AlertProxyRequest.FACTORY

        username, password = "Invalid auth-token submitted", ""
        auth_arr = auth.split(" ")
        if len(auth_arr) != 2:
            return

        atype, token = auth_arr
        if atype == "Basic":
            try:
                username, password = b64decode(token).split(":")
            except:  # noqa: E722
                pass
        elif atype == "NTLM":
            # b64decode returns bytes not str in python2
            print(b64decode(token).decode("utf-8").split(":"))
            exit(1)
            print("something NTLM")
            # Shouldn't this return something?
            return

        logdata = {"USERNAME": username, "PASSWORD": password}
        factory.log(logdata, transport=self.transport)

    def process(self):
        self.logAuth()

        factory = AlertProxyRequest.FACTORY
        profile = PROFILES[factory.skin]
        content = factory.auth_template.render(
            url=self.uri.decode(),
            date=datetime.datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S %ZGMT"),
            clientip=self.transport.getPeer().host,
        )
        if factory.banner:
            prompt = factory.banner
        else:
            prompt = profile.get("banner", "")

        #  for fooling nmap service detection
        if profile.get("HTTP1.1_always", False):
            self.clientproto = "HTTP/1.1"

        # match http-proxy m|^HTTP/1\.[01] \d\d\d .*\r\nServer: [sS]quid/([-.\w+]+)\r\n|s
        self.setResponseCode(407, profile["status_reason"])
        for (name, value) in profile["headers"]:
            self.responseHeaders.addRawHeader(name, value)

        self.responseHeaders.addRawHeader("Content-Type", "text/html")
        self.responseHeaders.addRawHeader(
            "Proxy-Authenticate", 'Basic realm="%s"' % prompt
        )
        self.responseHeaders.addRawHeader("Content-Length", "{}".format(len(content)))

        self.write(content.encode("utf-8"))
        self.finish()


class AlertProxy(HTTPChannel):
    requestFactory = AlertProxyRequest


class HTTPProxyFactory(http.HTTPFactory):
    def buildProtocol(self, addr):
        return AlertProxy()


class HTTPProxy(CanaryService):
    NAME = "httpproxy"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("httpproxy.port", default=8443))
        self.banner = config.getVal("httpproxy.banner", "").encode("utf8")
        self.skin = config.getVal("httpproxy.skin", default="squid")
        self.skindir = os.path.join(HTTPProxy.resource_dir(), "skin", self.skin)
        self.logtype = logger.LOG_HTTPPROXY_LOGIN_ATTEMPT
        self.listen_addr = config.getVal("device.listen_addr", default="")

        authfilename = os.path.join(self.skindir, "auth.html")
        try:
            with open(authfilename, "r") as f:
                self.auth_template = Template(f.read())
        except:  # noqa: E722
            self.auth_template = Template("")

    def getService(self):
        AlertProxyRequest.FACTORY = self
        f = HTTPProxyFactory()
        return internet.TCPServer(self.port, f, interface=self.listen_addr)



from opencanary.modules import CanaryService

from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet


class ProtocolError(Exception):
    pass


class GitCommandLengthMismatch(Exception):
    pass


class GitProtocol(Protocol):
    """
    Implementation of Git-daemon up to request
    """

    def _checkDataLength(self, data):
        try:
            actual_length = len(data)
            indata_length = int(data[0:4], base=16)
            if actual_length < indata_length:
                raise GitCommandLengthMismatch()
            elif actual_length > indata_length:
                raise ProtocolError()
        except ValueError:
            return False

    def _buildResponseAndSend(self, command):
        project = command[17 : 17 + command[17:].find("host")]
        request = command[command.find("=") + 1 :]
        self._logAlert(project, request)
        pre_response = "ERR no such repository: " + project
        response_size = "{:04x}".format(int(len(pre_response) + 4))
        response = response_size + pre_response
        self.transport.write(response.encode() + "\n".encode())

    def _logAlert(self, project, request):
        logdata = {"REPO": project[:-1], "HOST": request[:-1]}
        self.factory.log(logdata, transport=self.transport)

    def dataReceived(self, data):
        """
        Received data is unbuffered so we buffer it for telnet.
        """
        try:
            try:
                if not hasattr(self, "_data"):
                    self._data = data
                else:
                    self._data += data

                self._checkDataLength(self._data)

                git_command = self._data[4:]
                if git_command[:15] == b"git-upload-pack":
                    self._buildResponseAndSend(git_command.decode("utf-8"))
                else:
                    raise ProtocolError()

            except GitCommandLengthMismatch:
                pass

        except ProtocolError:
            self.transport.loseConnection()
            return


class CanaryGit(Factory, CanaryService):
    NAME = "git"
    protocol = GitProtocol

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = config.getVal("git.port", default=9418)
        self.listen_addr = config.getVal("device.listen_addr", default="")
        self.logtype = logger.LOG_GIT_CLONE_REQUEST

    def getService(self):
        return internet.TCPServer(self.port, self, interface=self.listen_addr)



"""
    A log-only SNMP server. It won't respond, but it will log SNMP queries.
"""
from opencanary.modules import CanaryService

from twisted.application import internet
from twisted.internet.protocol import DatagramProtocol

from twisted.internet.address import IPv4Address

from scapy.all import SNMP


class MiniSNMP(DatagramProtocol):
    def datagramReceived(self, data, host_and_port):
        try:
            snmp = SNMP(data)
            community = snmp.community.val
            requests = [x.oid.val for x in snmp.PDU.varbindlist]

            logdata = {"REQUESTS": requests, "COMMUNITY_STRING": community}
            self.transport.getPeer = lambda: IPv4Address(
                "UDP", host_and_port[0], host_and_port[1]
            )
            self.factory.log(logdata=logdata, transport=self.transport)
        except Exception as e:
            print(e)
            pass


class CanarySNMP(CanaryService):
    NAME = "SNMP"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("snmp.port", default=161))
        self.logtype = logger.LOG_SNMP_CMD
        self.listen_addr = config.getVal("device.listen_addr", default="")

    def getService(self):
        f = MiniSNMP()
        f.factory = self
        return internet.UDPServer(self.port, f, interface=self.listen_addr)



from opencanary.modules import CanaryService

from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet

from opencanary.modules.des import des

import os

RFB_33 = b"003.003"
RFB_37 = b"003.007"
RFB_38 = b"003.008"

# states
PRE_INIT = 1
HANDSHAKE_SEND = 2
SECURITY_SEND = 3
AUTH_SEND = 4
AUTH_OVER = 5

# if one of these is used in the VNC authentication attempt, alert that
# a common password was tried
COMMON_PASSWORDS = [
    "111111",
    "password",
    "123456",
    "111111",
    "1234",
    "administrator",
    "root",
    "passw0rd",
]


class ProtocolError(Exception):
    pass


class UnsupportedVersion(Exception):
    pass


class VNCProtocol(Protocol):
    """
    Implementation of VNC up to VNC authentication
    """

    def __init__(self, version=RFB_38):
        self.serv_version = version
        self.state = PRE_INIT

    def _send_handshake(
        self,
    ):
        print("send handshake")
        version_string = "RFB {version}\n".format(
            version=self.serv_version.decode("utf-8")
        )
        self.transport.write(version_string.encode("utf-8"))
        self.state = HANDSHAKE_SEND

    def _recv_handshake(self, data=None):
        print("got handshake")
        if len(data) != 12 or data[:3] != b"RFB":
            raise ProtocolError()
        client_ver = data[4:-1]

        # support single version for now
        if client_ver not in [RFB_33, RFB_37, RFB_38]:
            raise UnsupportedVersion()

        self._send_security(client_ver)

    def _send_security(self, client_ver):
        print("send security")
        if client_ver == RFB_33:
            self.transport.write(b"\x00\x00\x00\x02")  # specify VNC auth using 4 bytes
            self._send_auth()
        else:
            self.transport.write(b"\x01\x02")  # VNC authentication
            self.state = SECURITY_SEND

    def _recv_security(self, data=None):
        print("got security")
        if len(data) != 1 and data != "\x02":
            raise ProtocolError()
        self._send_auth()

    def _send_auth(
        self,
    ):
        print("send auth")
        self.challenge = os.urandom(16)
        self.transport.write(self.challenge)
        self.state = AUTH_SEND

    def _recv_auth(self, data=None):
        print("got auth")
        if len(data) != 16:
            raise ProtocolError()

        logdata = {
            "VNC Server Challenge": self.challenge.hex(),
            "VNC Client Response": data.hex(),
        }

        used_password = self._try_decrypt_response(response=data)
        if used_password:
            logdata["VNC Password"] = used_password
        else:
            logdata["VNC Password"] = "<Password was not in the common list>"
        self.factory.log(logdata, transport=self.transport)
        self._send_auth_failed()

    def connectionMade(self):
        if self.state != PRE_INIT:
            raise ProtocolError()
        self._send_handshake()

    def _send_auth_failed(
        self,
    ):
        self.transport.write(
            b"\x00\x00\x00\x01"
            + b"\x00\x00\x00\x16"  # response code
            + b"Authentication failure"  # message length
        )  # Message
        self.state = AUTH_OVER
        raise ProtocolError()

    def _try_decrypt_response(self, response=None):
        # attempt to decrypt each of the common passwords
        # really inefficient, but it means we don't have to rely on
        # a static challenge
        for password in COMMON_PASSWORDS:
            pw = password[:8]  # vnc passwords are max 8 chars
            if len(pw) < 8:
                pw += "\x00" * (8 - len(pw))

            pw = pw.encode("ascii")
            # VNC use of DES requires password bits to be mirrored
            values = bytearray()
            for x in pw:
                values.append(int("{:08b}".format(x)[::-1], 2))
            desbox = des(values)

            decrypted_challenge = desbox.decrypt(response)
            if decrypted_challenge == self.challenge:
                return password
        return None

    def dataReceived(self, data):
        """
        Received data is unbuffered so we buffer it for telnet.
        """
        try:
            if self.state == HANDSHAKE_SEND:
                self._recv_handshake(data=data)
            elif self.state == SECURITY_SEND:
                self._recv_security(data=data)
            elif self.state == AUTH_SEND:
                self._recv_auth(data=data)
        except (UnsupportedVersion, ProtocolError):
            self.transport.loseConnection()
            return


class CanaryVNC(Factory, CanaryService):
    NAME = "VNC"
    protocol = VNCProtocol

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config, logger)
        self.port = config.getVal("vnc.port", 5900)
        self.listen_addr = config.getVal("device.listen_addr", default="")
        self.logtype = logger.LOG_VNC

    def getService(self):
        return internet.TCPServer(self.port, self, interface=self.listen_addr)


CanaryServiceFactory = CanaryVNC



from opencanary.modules import CanaryService
from opencanary.modules import FileSystemWatcher
import os
import shutil
import subprocess


class SynLogWatcher(FileSystemWatcher):
    def __init__(
        self, logger=None, logFile=None, ignore_localhost=False, ignore_ports=None
    ):
        if ignore_ports is None:
            ignore_ports = []
        self.logger = logger
        self.ignore_localhost = ignore_localhost
        self.ignore_ports = ignore_ports
        FileSystemWatcher.__init__(self, fileName=logFile)

    def handleLines(self, lines=None):  # noqa: C901
        for line in lines:
            try:
                if "canaryfw: " in line:
                    logtype = self.logger.LOG_PORT_SYN
                    (rubbish, log) = line.split("canaryfw: ")
                elif "canarynmapNULL" in line:
                    logtype = self.logger.LOG_PORT_NMAPNULL
                    (rubbish, log) = line.split("canarynmapNULL: ")
                elif "canarynmapXMAS" in line:
                    logtype = self.logger.LOG_PORT_NMAPXMAS
                    (rubbish, log) = line.split("canarynmapXMAS: ")
                elif "canarynmapFIN" in line:
                    logtype = self.logger.LOG_PORT_NMAPFIN
                    (rubbish, log) = line.split("canarynmapFIN: ")
                elif "canarynmap: " in line:
                    logtype = self.logger.LOG_PORT_NMAPOS
                    (rubbish, log) = line.split("canarynmap: ")
                else:
                    continue
            except ValueError:
                continue
            tags = log.split(" ")
            kv = {}
            for tag in tags:
                if tag.find("=") >= 0:
                    (key, val) = tag.split("=")
                else:
                    key = tag
                    val = ""
                kv[key] = val

            # we've seen empty tags creep in. weed them out.
            if "" in kv.keys():
                kv.pop("")

            data = {}
            data["src_host"] = kv.pop("SRC")
            data["src_port"] = kv.pop("SPT")
            data["dst_host"] = kv.pop("DST")
            data["dst_port"] = kv.pop("DPT")
            data["logtype"] = logtype
            data["logdata"] = kv
            if self.ignore_localhost and data.get("src_host", False) == "127.0.0.1":
                continue
            if int(data.get("dst_port", -1)) in self.ignore_ports:
                continue

            self.logger.log(data)


def detectNFTables():
    return b"nf_tables" in subprocess.check_output(["iptables", "--version"])


class CanaryPortscan(CanaryService):
    NAME = "portscan"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.audit_file = config.getVal("portscan.logfile", default="/var/log/kern.log")
        self.synrate = int(config.getVal("portscan.synrate", default=5))
        self.nmaposrate = int(config.getVal("portscan.nmaposrate", default="5"))
        self.lorate = int(config.getVal("portscan.lorate", default="3"))
        self.listen_addr = config.getVal("device.listen_addr", default="")
        self.ignore_localhost = config.getVal(
            "portscan.ignore_localhost", default=False
        )
        self.ignore_ports = config.getVal("portscan.ignore_ports", default=[])
        self.iptables_path = self.config.getVal("portscan.iptables_path", False)
        self.config = config

    def getIptablesPath(self):
        if self.iptables_path:
            return self.iptables_path

        if detectNFTables():
            return shutil.which("iptables-legacy")

        return shutil.which("iptables") or "/sbin/iptables"

    def startYourEngines(self, reactor=None):
        # Logging rules for loopback interface.
        # This is separate from the canaryfw rule as the canary watchdog was
        # causing console-side noise in the logs.
        self.set_iptables_rules()

        fs = SynLogWatcher(
            logFile=self.audit_file,
            logger=self.logger,
            ignore_localhost=self.ignore_localhost,
            ignore_ports=self.ignore_ports,
        )
        fs.start()

    def configUpdated(
        self,
    ):
        pass

    def set_iptables_rules(self):
        iptables_path = self.getIptablesPath()
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp -i lo -j LOG --log-level=warning --log-prefix="canaryfw: " -m limit --limit="{1}/hour"'.format(
                iptables_path, self.lorate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp -i lo -j LOG --log-level=warning --log-prefix="canaryfw: " -m limit --limit="{1}/hour"'.format(
                iptables_path, self.lorate
            )
        )

        # Logging rules for canaryfw.
        # We ignore loopback interface traffic as it is taken care of in above rule
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp --syn -j LOG --log-level=warning --log-prefix="canaryfw: " -m limit --limit="{1}/second" ! -i lo'.format(
                iptables_path, self.synrate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp --syn -j LOG --log-level=warning --log-prefix="canaryfw: " -m limit --limit="{1}/second" ! -i lo'.format(
                iptables_path, self.synrate
            )
        )

        # Match the T3 probe of the nmap OS detection based on TCP flags and TCP options string
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp --tcp-flags ALL URG,PSH,SYN,FIN -m u32 --u32 "40=0x03030A01 && 44=0x02040109 && 48=0x080Affff && 52=0xffff0000 && 56=0x00000402" -j LOG --log-level=warning --log-prefix="canarynmap: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp --tcp-flags ALL URG,PSH,SYN,FIN -m u32 --u32 "40=0x03030A01 && 44=0x02040109 && 48=0x080Affff && 52=0xffff0000 && 56=0x00000402" -j LOG --log-level=warning --log-prefix="canarynmap: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )

        # Nmap Null Scan
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50000400" -j LOG --log-level=warning --log-prefix="canarynmapNULL: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50000400" -j LOG --log-level=warning --log-prefix="canarynmapNULL: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )

        # Nmap Xmas Scan
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50290400" -j LOG --log-level=warning --log-prefix="canarynmapXMAS: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50290400" -j LOG --log-level=warning --log-prefix="canarynmapXMAS: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )

        # Nmap Fin Scan
        os.system(
            'sudo {0} -t mangle -D PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50010400" -j LOG --log-level=warning --log-prefix="canarynmapFIN: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )
        os.system(
            'sudo {0} -t mangle -A PREROUTING -p tcp -m u32 --u32 "6&0xFF=0x6 && 0>>22&0x3C@12=0x50010400" -j LOG --log-level=warning --log-prefix="canarynmapFIN: " -m limit --limit="{1}/second"'.format(
                iptables_path, self.nmaposrate
            )
        )



from opencanary.modules import CanaryService
import socket

from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet


class ProtocolError(Exception):
    pass


class UnsupportedVersion(Exception):
    pass


class TCPBannerProtocol(Protocol):
    """
    Implementation of TCP Banner module - reply with a text banner
    """

    def __init__(
        self,
        factory,
        banner_id,
        accept_banner,
        send_banner,
        alert_string_enabled,
        alert_string,
        keep_alive_enabled,
        keep_alive_secret,
        keep_alive_idle,
        keep_alive_interval,
        keep_alive_probes,
    ):
        self.factory = factory
        self.banner_id = banner_id
        self.accept_banner = accept_banner
        self.send_banner = send_banner
        self.alert_string_enabled = alert_string_enabled
        self.alert_string = alert_string
        self.keep_alive_enabled = keep_alive_enabled
        # once we send the secret key we disable alerting on keep-alive
        # connections
        self.keep_alive_disable_alerting = False
        self.keep_alive_secret = keep_alive_secret
        self.keep_alive_idle = keep_alive_idle
        self.keep_alive_interval = keep_alive_interval
        self.keep_alive_probes = keep_alive_probes

    def connectionMade(self):
        # We limit the data sent through to 255 chars
        try:
            data = str(self.accept_banner)[:255]

            logdata = {
                "FUNCTION": "CONNECTION_MADE",
                "DATA": data,
                "BANNER_ID": str(self.banner_id),
            }

            if self.keep_alive_enabled:
                if hasattr(socket, "TCP_KEEPIDLE"):
                    # overrides value (in seconds) of system-wide ipv4 tcp_keepalive_time
                    self.transport.getHandle().setsockopt(
                        socket.SOL_TCP, socket.TCP_KEEPIDLE, self.keep_alive_idle
                    )
                if hasattr(socket, "TCP_KEEPINTVL"):
                    # overrides value (in seconds) of system-wide ipv4 tcp_keepalive_intvl
                    self.transport.getHandle().setsockopt(
                        socket.SOL_TCP,
                        socket.TCP_KEEPINTVL,
                        int(self.keep_alive_interval),
                    )
                if hasattr(socket, "TCP_KEEPCNT"):
                    # overrides value (in seconds) of system-wide ipv4 tcp_keepalive_probes
                    self.transport.getHandle().setsockopt(
                        socket.SOL_TCP, socket.TCP_KEEPCNT, self.keep_alive_probes
                    )
                # set keep alive on socket
                self.transport.setTcpKeepAlive(1)

                self.factory.canaryservice.logtype = (
                    self.factory.canaryservice.logger.LOG_TCP_BANNER_KEEP_ALIVE_CONNECTION_MADE
                )
                self.factory.canaryservice.log(logdata, transport=self.transport)

            elif not self.alert_string_enabled:
                # flag says we need to wait for incoming data to include a string
                # so no point in logging anything here

                self.factory.canaryservice.logtype = (
                    self.factory.canaryservice.logger.LOG_TCP_BANNER_CONNECTION_MADE
                )
                self.factory.canaryservice.log(logdata, transport=self.transport)

            self.transport.write(self.accept_banner)

        except OSError:
            print("Received an OSError. Likely the socket has closed.")
            self.factory.canaryservice.logtype = (
                self.factory.canaryservice.logger.LOG_TCP_BANNER_CONNECTION_MADE
            )
            self.factory.canaryservice.log(logdata, transport=self.transport)

    def dataReceived(self, data):
        """
        Received data from tcp connection after connection has been made.
        """
        try:
            if self.keep_alive_disable_alerting:
                self.transport.write(self.send_banner)
                return

            # We limit the data sent through to 255 chars
            data = data[:255]

            logdata = {"FUNCTION": "DATA_RECEIVED", "BANNER_ID": str(self.banner_id)}
            try:
                logdata["DATA"] = data.rstrip().decode().encode("utf-8")
            except UnicodeDecodeError:
                logdata["DATA"] = data.rstrip().decode("unicode_escape").encode("utf-8")

            if self.keep_alive_enabled:
                self.keep_alive_action(data, logdata)
            else:
                self.non_keep_alive_action(data, logdata)

            self.transport.write(self.send_banner)
        except (UnsupportedVersion, ProtocolError):
            self.transport.loseConnection()
            return

    def non_keep_alive_action(self, data, logdata):
        self.factory.canaryservice.logtype = (
            self.factory.canaryservice.logger.LOG_TCP_BANNER_DATA_RECEIVED
        )
        if self.alert_string_enabled:
            if self.alert_string in data:
                logdata["ALERT_STRING"] = (self.alert_string).decode().encode("utf-8")
                self.factory.canaryservice.log(logdata, transport=self.transport)

    def keep_alive_action(self, data, logdata):
        if self.keep_alive_secret != "" and self.keep_alive_secret in data:
            self.keep_alive_disable_alerting = True
            self.factory.canaryservice.logtype = (
                self.factory.canaryservice.logger.LOG_TCP_BANNER_KEEP_ALIVE_SECRET_RECEIVED
            )
            logdata["SECRET_STRING"] = (self.keep_alive_secret).decode().encode("utf-8")
        else:
            self.factory.canaryservice.logtype = (
                self.factory.canaryservice.logger.LOG_TCP_BANNER_KEEP_ALIVE_DATA_RECEIVED
            )
        self.factory.canaryservice.log(logdata, transport=self.transport)


class TCPBannerFactory(Factory):
    def __init__(self, config=None, banner_id=1):
        self.banner_id = str(banner_id)
        self.accept_banner = (
            config.getVal("tcpbanner_" + self.banner_id + ".initbanner", "")
            .encode("utf8")
            .replace(b"\\n", b"\n")
            .replace(b"\\r", b"\r")
        )
        self.send_banner = (
            config.getVal("tcpbanner_" + self.banner_id + ".datareceivedbanner", "")
            .encode("utf8")
            .replace(b"\\n", b"\n")
            .replace(b"\\r", b"\r")
        )
        self.alert_string_enabled = config.getVal(
            "tcpbanner_" + self.banner_id + ".alertstring.enabled", False
        )
        self.alert_string = (
            config.getVal("tcpbanner_" + self.banner_id + ".alertstring", "")
            .encode("utf8")
            .replace(b"\\n", b"\n")
            .replace(b"\\r", b"\r")
        )
        self.keep_alive_enabled = config.getVal(
            "tcpbanner_" + self.banner_id + ".keep_alive.enabled", False
        )
        self.keep_alive_secret = (
            config.getVal("tcpbanner_" + self.banner_id + ".keep_alive_secret", "")
            .encode("utf8")
            .replace(b"\\n", b"\n")
            .replace(b"\\r", b"\r")
        )
        # the defaults for the tcp keep alive values add up to 1 hour keep alive
        # 300 for the first probe + 300 * 11 for the interval probes = 3600
        self.keep_alive_idle = config.getVal(
            "tcpbanner_" + self.banner_id + ".keep_alive_idle", 300
        )
        self.keep_alive_interval = config.getVal(
            "tcpbanner_" + self.banner_id + ".keep_alive_interval", 300
        )
        self.keep_alive_probes = config.getVal(
            "tcpbanner_" + self.banner_id + ".keep_alive_probes", 11
        )

    def buildProtocol(self, addr):
        return TCPBannerProtocol(
            self,
            self.banner_id,
            self.accept_banner,
            self.send_banner,
            self.alert_string_enabled,
            self.alert_string,
            self.keep_alive_enabled,
            self.keep_alive_secret,
            self.keep_alive_idle,
            self.keep_alive_interval,
            self.keep_alive_probes,
        )


class CanaryTCPBanner(CanaryService):
    NAME = "tcpbanner"
    MAX_TCP_BANNERS = 10

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)

    def getService(self):
        services = []
        for i in range(
            1, self.config.getVal("tcpbanner.maxnum", default=self.MAX_TCP_BANNERS) + 1
        ):
            if self.config.getVal("tcpbanner_" + str(i) + ".enabled", False):
                factory = TCPBannerFactory(config=self.config, banner_id=i)
                factory.canaryservice = self
                port = self.config.getVal(
                    "tcpbanner_" + str(i) + ".port", default=8000 + i
                )
                services.append(internet.TCPServer(port, factory))
        return services



import struct
import socket


def ip2int(addr):
    """
    Convert an IP in string format to decimal format
    """

    return struct.unpack("!I", socket.inet_aton(addr))[0]


def check_ip(ip, network_range):
    """
    Test if the IP is in range

    Range is expected to be in CIDR notation format. If no MASK is
    given /32 is used. It return True if the IP is in the range.
    """

    netItem = str(network_range).split("/")
    rangeIP = netItem[0]
    if len(netItem) == 2:
        rangeMask = int(netItem[1])
    else:
        rangeMask = 32

    try:
        ripInt = ip2int(rangeIP)
        ipInt = ip2int(ip)
        result = not ((ipInt ^ ripInt) & 0xFFFFFFFF << (32 - rangeMask))
    except:  # noqa: E722
        result = False

    return result



import re

from opencanary.modules import CanaryService

from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet


class RemoteDesktopProtocol(Protocol):
    """
    A simple service that logs RDP connection attempts
    and mimics an NLA-enabled RDP server but responds with login failure
    """

    def __init__(self):
        self.initial_connection = True

    def dataReceived(self, data):
        # Decode the data to unicode, so we can search it, and ignore any errors
        # caused by bytes that can't be decoded.
        decoded_data = data.decode(encoding="utf-8", errors="ignore")
        # Use regex to extract the username.
        match = re.search(r"mstshash=(?P<username>[a-zA-Z0-9-_@]*)", decoded_data)
        username = match and match.groupdict().get("username")
        # Log the connection attempt
        self.factory.log(logdata={"USERNAME": username}, transport=self.transport)

        if self.initial_connection:
            # Respond as an NLA-enabled RDP server
            self.transport.write(
                bytes.fromhex("030000130ed000001234000209080002000000")
            )
            self.initial_connection = False
        else:
            # Always respond with a negotiation failure, details from
            # https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/96327ab4-d43f-4803-9aff-392ce1fc2073
            self.transport.write(bytes.fromhex("0001000400010000052e"))
            self.transport.loseConnection()


class CanaryRDP(Factory, CanaryService):
    NAME = "rdp"
    protocol = RemoteDesktopProtocol

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config, logger)
        self.port = config.getVal("rdp.port", 3389)
        self.listen_addr = config.getVal("device.listen_addr", default="")
        self.logtype = logger.LOG_RDP

    def getService(self):
        return internet.TCPServer(self.port, self, interface=self.listen_addr)


CanaryServiceFactory = CanaryRDP



#!/usr/bin/env python3
# Python script to generate a .plist file that launchctl can use to manage opencanary as a service
# as well as bootstrap and bootout scripts to get the service up and running.
# NOTE: Requires homebrew.

import json
import pathlib
import plistlib
import re
import stat
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
from os import chmod, pardir, path
from os.path import dirname, join, realpath
from subprocess import CalledProcessError, check_output

from pkg_resources import resource_filename

OPENCANARY = "opencanary"
LAUNCH_DAEMONS_DIR = "/Library/LaunchDaemons"
DEFAULT_SERVICE_NAME = "com.thinkst.opencanary"
CONFIG_FILE_BASENAME = "opencanary.conf"
DEFAULT_CONFIG_FILE = resource_filename("opencanary", "data/settings.json")

# opencanary dirs
OPENCANARY_DIR = realpath(join(dirname(__file__), pardir))
OPENCANARY_BIN_DIR = join(OPENCANARY_DIR, "bin")
VENV_DIR = join(OPENCANARY_DIR, "env")
VENV_BIN_DIR = join(VENV_DIR, "bin")
DEFAULT_LOG_DIR = join(OPENCANARY_DIR, "log")

# daemon config
DAEMON_CONFIG_DIR = "/etc/opencanaryd"
DAEMON_CONFIG_PATH = join(DAEMON_CONFIG_DIR, CONFIG_FILE_BASENAME)
DAEMON_PATH = join(VENV_BIN_DIR, "opencanaryd")
DAEMON_RUNTIME_OPTIONS = "--dev"

# This script writes to the launchctl/ folder
LAUNCHCTL_DIR = join(OPENCANARY_DIR, "launchctl")

# Homebrew (TODO: is this necessary?)
try:
    homebrew_bin_dir = join(check_output(["brew", "--prefix"]).decode().rstrip(), "bin")
except CalledProcessError as e:
    print(f"Couldn't get homebrew install location: {e}")
    sys.exit()

# Load opencanary.conf default config
with open(DEFAULT_CONFIG_FILE, "r") as file:
    config = json.load(file)
    canaries = [
        k.split(".")[0] for k in config.keys() if re.match("[a-z]+\\.enabled", k)
    ]


# Parse arguments.
parser = ArgumentParser(
    description="Generate .plist, opencanary.conf, and scripts to bootstrap opencanary as a launchctl daemon.",
    formatter_class=ArgumentDefaultsHelpFormatter,
)

parser.add_argument(
    "--service-name",
    help="string you would like launchctl to use as the name of the opencanary service",
    metavar="NAME",
    default=DEFAULT_SERVICE_NAME,
)

parser.add_argument(
    "--log-output-dir",
    help="opencanary will write its logs to files in DIR when the service is running",
    metavar="DIR",
    default=DEFAULT_LOG_DIR,
)

parser.add_argument(
    "--canary",
    action="append",
    help="enable canary service in the generated opencanary.conf file "
    + "(can be supplied more than once)",
    choices=canaries,
    dest="canaries",
)


args = parser.parse_args()
args.canaries = args.canaries or []
plist_basename = args.service_name + ".plist"

# Setup dirs
for dir in [LAUNCHCTL_DIR, args.log_output_dir]:
    print(f"Creating '{dir}'...")
    pathlib.Path(dir).mkdir(parents=True, exist_ok=True)

# File builders
build_launchctl_dir_path = partial(join, LAUNCHCTL_DIR)
build_logfile_name = lambda log_name: join(  # noqa: E731
    args.log_output_dir, f"opencanary.{log_name}.log"
)


# daemon launcher script
launcher_script = build_launchctl_dir_path(f"launch_{args.service_name}.sh")

with open(launcher_script, "w") as file:
    file.write(f'. "{VENV_BIN_DIR}/activate"\n')
    file.write(f'"{DAEMON_PATH}" {DAEMON_RUNTIME_OPTIONS}\n')


# Write launchctl service .plist
plist_output_file = build_launchctl_dir_path(plist_basename)

plist_contents = {
    "Label": args.service_name,
    "RunAtLoad": True,
    "KeepAlive": True,
    "WorkingDirectory": VENV_BIN_DIR,
    "StandardOutPath": build_logfile_name("err"),
    "StandardErrorPath": build_logfile_name("out"),
    "EnvironmentVariables": {
        "PATH": f"{VENV_BIN_DIR}:{homebrew_bin_dir}:/usr/bin:/bin",
        "VIRTUAL_ENV": VENV_DIR,
    },
    "ProgramArguments": [launcher_script],
}

with open(plist_output_file, "wb+") as _plist_file:
    plistlib.dump(plist_contents, _plist_file)


# opencanary config
for canary in canaries:
    config[f"{canary}.enabled"] = canary in args.canaries

log_handlers = config["logger"]["kwargs"]["handlers"]
log_handlers["file"]["filename"] = build_logfile_name("run")

# TODO: This config doesn't work even though direct calls to syslog do
# log_handlers["syslog-unix"] = {
#     "class": "logging.handlers.SysLogHandler",
#     "formatter":"syslog_rfc",
#     "address": [
#         "localhost",
#         514
#     ],
#     "socktype": "ext://socket.SOCK_DGRAM"
# }

config_output_file = build_launchctl_dir_path(CONFIG_FILE_BASENAME)

with open(config_output_file, "w") as file:
    file.write(json.dumps(config, indent=4))


# service bootstrap script
install_service_script = build_launchctl_dir_path(
    f"install_service_{args.service_name}.sh"
)
daemon_plist_path = join(LAUNCH_DAEMONS_DIR, plist_basename)

with open(install_service_script, "w") as file:
    script_contents = [
        f"set -e\n\n" f"chown root '{launcher_script}'",
        f"mkdir -p '{DAEMON_CONFIG_DIR}'",
        f"cp '{config_output_file}' {DAEMON_CONFIG_PATH}",
        f"cp '{plist_output_file}' {LAUNCH_DAEMONS_DIR}",
        f"launchctl bootstrap system '{daemon_plist_path}'",
        "",
    ]

    file.write("\n".join(script_contents))


# uninstall/bootout script
uninstall_service_script = build_launchctl_dir_path(
    f"uninstall_service_{args.service_name}.sh"
)

with open(uninstall_service_script, "w") as file:
    file.write(f"launchctl bootout system/{args.service_name}\n")


# Set permissions
chmod(launcher_script, stat.S_IRWXU)  # stat.S_IEXEC | stat.S_IREAD
chmod(uninstall_service_script, stat.S_IRWXU)
chmod(install_service_script, stat.S_IRWXU)


# Print results
print("Generated files...\n")
print(f"    Service definition: ./{path.relpath(plist_output_file)}")
print(f"       Launcher script: ./{path.relpath(launcher_script)}")
print(f"      Bootstrap script: ./{path.relpath(install_service_script)}")
print(f"        Bootout script: ./{path.relpath(uninstall_service_script)}\n")
print(f"                Config: ./{path.relpath(config_output_file)}")
print(f"      Enabled canaries: {', '.join(args.canaries)}\n")
print(f"To install as a system service run:\n    'sudo {install_service_script}'\n")



import os
import sys
import json
import itertools
import string
import subprocess
import shutil
import re
from os.path import expanduser
from pkg_resources import resource_filename
from pathlib import Path

SAMPLE_SETTINGS = resource_filename(__name__, "data/settings.json")
SETTINGS = "opencanary.conf"


def expand_vars(var):
    """Recursively replace environment variables in a dictionary, list or string with their respective values."""
    if isinstance(var, dict):
        for key, value in var.items():
            var[key] = expand_vars(value)
        return var
    if isinstance(var, (list, set, tuple)):
        return [expand_vars(v) for v in var]
    if isinstance(var, (str, bytes)):
        return os.path.expandvars(var)
    return var


def is_docker():
    cgroup = Path("/proc/self/cgroup")
    return (
        Path("/.dockerenv").is_file()
        or cgroup.is_file()
        and "docker" in cgroup.read_text()
    )


def detectIPTables():
    if shutil.which("iptables"):
        return True
    else:
        return False


SERVICE_REGEXES = {
    "ssh.version": r"(SSH-(2.0|1.5|1.99|1.0)-([!-,\-./0-~]+(:?$|\s))(?:[ -~]*)){1,253}$",
}


class Config:
    def __init__(self, configfile=SETTINGS):
        self.__config = None
        self.__configfile = configfile

        files = [
            configfile,
            "%s/.%s" % (expanduser("~"), configfile),
            "/etc/opencanaryd/%s" % configfile,
        ]
        print(
            "** We hope you enjoy using OpenCanary. For more open source Canary goodness, head over to canarytokens.org. **"
        )
        for fname in files:
            try:
                with open(fname, "r") as f:
                    print("[-] Using config file: %s" % fname)
                    self.__config = json.load(f)
                    self.__config = expand_vars(self.__config)
                return
            except IOError as e:
                print("[-] Failed to open %s for reading (%s)" % (fname, e))
            except ValueError as e:
                print("[-] Failed to decode json from %s (%s)" % (fname, e))
                subprocess.call(
                    "cp -r %s /var/tmp/config-err-$(date +%%s)" % fname, shell=True
                )
            except Exception as e:
                print("[-] An error occurred loading %s (%s)" % (fname, e))
        if self.__config is None:
            print(
                'No config file found. Please create one with "opencanaryd --copyconfig"'
            )
            sys.exit(1)

    def moduleEnabled(self, module_name):
        k = "%s.enabled" % module_name.lower()
        if k in self.__config:
            return bool(self.__config[k])
        return False

    def getVal(self, key, default=None):
        # throw exception to caller
        try:
            return self.__config[key]
        except KeyError as e:
            if default is not None:
                return default
            raise e

    def checkValues(self):  # noqa: C901
        """Set all the valid values in params and return a list of errors for invalid"""
        params = self.__config
        # test options indpenedently for validity
        errors = []
        for key, value in params.items():
            try:
                self.is_valid(key, value)
            except ConfigException as e:
                errors.append(e)

        # Test that no ports overlap
        ports = {k: int(v) for k, v in params.items() if k.endswith(".port")}
        ports = [(port, setting) for setting, port in ports.items()]
        ports.sort()

        for port, settings in itertools.groupby(ports, lambda x: x[0]):
            settings = list(settings)
            if len(settings) > 1:
                services = ", ".join([s[1].split(".")[0] for s in settings])
                errmsg = "More than one service uses this port (%s)" % services
                for (port, setting) in settings:
                    errors.append(ConfigException(setting, errmsg))

        return errors

    def is_valid(self, key, val):  # noqa: C901
        """
        Test an the validity of an individual setting
        Raise config error message on failure.
        TODO: delegate module tests to appropriate module
        """

        if key.endswith(".enabled"):
            if not ((val is True) or (val is False)):
                raise ConfigException(
                    key, "Boolean setting is not True or False (%s)" % val
                )

        if key.endswith(".port"):
            if not isinstance(val, int):
                raise ConfigException(
                    key, "Invalid port number (%s). Must be an integer." % val
                )
            if val < 1 or val > 65535:
                raise ConfigException(
                    key, "Invalid port number (%s). Must be between 1 and 65535." % val
                )
        # Max length of SSH version string is 255 chars including trailing CR and LF
        # https://tools.ietf.org/html/rfc4253
        if key == "ssh.version" and len(val) > 253:
            raise ConfigException(key, "SSH version string too long (%s..)" % val[:5])

        if key == "device.name":
            allowed_chars = string.ascii_letters + string.digits + "+-#_"

            if len(val) > 100:
                raise ConfigException(key, "Name cannot be longer than 100 characters")
            elif len(val) < 1:
                raise ConfigException(key, "Name ought to be at least one character")
            elif any(map(lambda x: x not in allowed_chars, val)):
                raise ConfigException(
                    key,
                    "Please use only characters, digits, any of the following: + - # _",
                )

        if key == "device.desc":
            allowed_chars = string.ascii_letters + string.digits + "+-#_ "
            if len(val) > 100:
                raise ConfigException(key, "Name cannot be longer than 100 characters")
            elif len(val) < 1:
                raise ConfigException(key, "Name ought to be at least one character")
            elif any(map(lambda x: x not in allowed_chars, val)):
                raise ConfigException(
                    key,
                    "Please use only characters, digits, spaces and any of the following: + - # _",
                )

        if key in SERVICE_REGEXES.keys():
            if not re.match(SERVICE_REGEXES[key], val):
                raise ConfigException(key, f"{val} is not valid.")

        return True

    def __repr__(self):
        return self.__config.__repr__()

    def __str__(self):
        return self.__config.__str__()

    def toDict(self):
        """Return all settings as a dict"""
        return self.__config

    def toJSON(self):
        """
        JSON representation of config
        """
        return json.dumps(
            self.__config, sort_keys=True, indent=4, separators=(",", ": ")
        )


class ConfigException(Exception):
    """Exception raised on invalid config value"""

    def __init__(self, key, msg):
        self.key = key
        self.msg = msg

    def __str__(self):
        return "%s: %s" % (self.key, self.msg)

    def __repr__(self):
        return "<%s %s (%s)>" % (self.__class__.__name__, self.key, self.msg)


config = Config()
errors = config.checkValues()
if errors:
    for error in errors:
        print(error)
    sys.exit(1)



#############################################################################
#               Documentation                   #
#############################################################################

# Author:   Todd Whiteman
# Date:     16th March, 2009
# Verion:   2.0.0
# License:  Public Domain - free to do as you wish
# Homepage: http://twhiteman.netfirms.com/des.html
#
# This is a pure python implementation of the DES encryption algorithm.
# It's pure python to avoid portability issues, since most DES
# implementations are programmed in C (for performance reasons).
#
# Triple DES class is also implemented, utilising the DES base. Triple DES
# is either DES-EDE3 with a 24 byte key, or DES-EDE2 with a 16 byte key.
#
# See the README.txt that should come with this python module for the
# implementation methods used.
#
# Thanks to:
#  * David Broadwell for ideas, comments and suggestions.
#  * Mario Wolff for pointing out and debugging some triple des CBC errors.
#  * Santiago Palladino for providing the PKCS5 padding technique.
#  * Shaya for correcting the PAD_PKCS5 triple des CBC errors.
#
"""A pure python implementation of the DES and TRIPLE DES encryption algorithms.

Class initialization
--------------------
pyDes.des(key, [mode], [IV], [pad], [padmode])
pyDes.triple_des(key, [mode], [IV], [pad], [padmode])

key     -> Bytes containing the encryption key. 8 bytes for DES, 16 or 24 bytes
       for Triple DES
mode    -> Optional argument for encryption type, can be either
       pyDes.ECB (Electronic Code Book) or pyDes.CBC (Cypher Block Chaining)
IV      -> Optional Initial Value bytes, must be supplied if using CBC mode.
       Length must be 8 bytes.
pad     -> Optional argument, set the pad character (PAD_NORMAL) to use during
       all encrypt/decrpt operations done with this instance.
padmode -> Optional argument, set the padding mode (PAD_NORMAL or PAD_PKCS5)
       to use during all encrypt/decrpt operations done with this instance.

I recommend to use PAD_PKCS5 padding, as then you never need to worry about any
padding issues, as the padding can be removed unambiguously upon decrypting
data that was encrypted using PAD_PKCS5 padmode.

Common methods
--------------
encrypt(data, [pad], [padmode])
decrypt(data, [pad], [padmode])

data    -> Bytes to be encrypted/decrypted
pad     -> Optional argument. Only when using padmode of PAD_NORMAL. For
       encryption, adds this characters to the end of the data block when
       data is not a multiple of 8 bytes. For decryption, will remove the
       trailing characters that match this pad character from the last 8
       bytes of the unencrypted data block.
padmode -> Optional argument, set the padding mode, must be one of PAD_NORMAL
       or PAD_PKCS5). Defaults to PAD_NORMAL.


Example
-------
from pyDes import *

data = "Please encrypt my data"
k = des("DESCRYPT", CBC, "\0\0\0\0\0\0\0\0", pad=None, padmode=PAD_PKCS5)
# For Python3, you'll need to use bytes, i.e.:
#   data = b"Please encrypt my data"
#   k = des(b"DESCRYPT", CBC, b"\0\0\0\0\0\0\0\0", pad=None, padmode=PAD_PKCS5)
d = k.encrypt(data)
print"Encrypted: %r" % d
print "Decrypted: %r" % k.decrypt(d)
assert k.decrypt(d, padmode=PAD_PKCS5) == data


See the module source (pyDes.py) for more examples of use.
You can also run the pyDes.py file without and arguments to see a simple test.

Note: This code was not written for high-end systems needing a fast
      implementation, but rather a handy portable solution with small usage.

"""

import sys

# _pythonMajorVersion is used to handle Python2 and Python3 differences.
_pythonMajorVersion = sys.version_info[0]

# Modes of crypting / cyphering
ECB = 0
CBC = 1

# Modes of padding
PAD_NORMAL = 1
PAD_PKCS5 = 2

# PAD_PKCS5: is a method that will unambiguously remove all padding
#            characters after decryption, when originally encrypted with
#            this padding mode.
# For a good description of the PKCS5 padding technique, see:
# http://www.faqs.org/rfcs/rfc1423.html

# The base class shared by des and triple des.
class _baseDes(object):
    def __init__(self, mode=ECB, IV=None, pad=None, padmode=PAD_NORMAL):
        if IV:
            IV = self._guardAgainstUnicode(IV)
        if pad:
            pad = self._guardAgainstUnicode(pad)
        self.block_size = 8
        # Sanity checking of arguments.
        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")
        if IV and len(IV) != self.block_size:
            raise ValueError(
                "Invalid Initial Value (IV), must be a multiple of "
                + str(self.block_size)
                + " bytes"
            )

        # Set the passed in variables
        self._mode = mode
        self._iv = IV
        self._padding = pad
        self._padmode = padmode

    def getKey(self):
        """getKey() -> bytes"""
        return self.__key

    def setKey(self, key):
        """Will set the crypting key for this object."""
        key = self._guardAgainstUnicode(key)
        self.__key = key

    def getMode(self):
        """getMode() -> pyDes.ECB or pyDes.CBC"""
        return self._mode

    def setMode(self, mode):
        """Sets the type of crypting mode, pyDes.ECB or pyDes.CBC"""
        self._mode = mode

    def getPadding(self):
        """getPadding() -> bytes of length 1. Padding character."""
        return self._padding

    def setPadding(self, pad):
        """setPadding() -> bytes of length 1. Padding character."""
        if pad is not None:
            pad = self._guardAgainstUnicode(pad)
        self._padding = pad

    def getPadMode(self):
        """getPadMode() -> pyDes.PAD_NORMAL or pyDes.PAD_PKCS5"""
        return self._padmode

    def setPadMode(self, mode):
        """Sets the type of padding mode, pyDes.PAD_NORMAL or pyDes.PAD_PKCS5"""
        self._padmode = mode

    def getIV(self):
        """getIV() -> bytes"""
        return self._iv

    def setIV(self, IV):
        """Will set the Initial Value, used in conjunction with CBC mode"""
        if not IV or len(IV) != self.block_size:
            raise ValueError(
                "Invalid Initial Value (IV), must be a multiple of "
                + str(self.block_size)
                + " bytes"
            )
        IV = self._guardAgainstUnicode(IV)
        self._iv = IV

    def _padData(self, data, pad, padmode):
        # Pad data depending on the mode
        if padmode is None:
            # Get the default padding mode.
            padmode = self.getPadMode()
        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")

        if padmode == PAD_NORMAL:
            if len(data) % self.block_size == 0:
                # No padding required.
                return data

            if not pad:
                # Get the default padding.
                pad = self.getPadding()
            if not pad:
                raise ValueError(
                    "Data must be a multiple of "
                    + str(self.block_size)
                    + " bytes in length. Use padmode=PAD_PKCS5 or set the pad character."
                )
            data += (self.block_size - (len(data) % self.block_size)) * pad

        elif padmode == PAD_PKCS5:
            pad_len = 8 - (len(data) % self.block_size)
            if _pythonMajorVersion < 3:
                data += pad_len * chr(pad_len)
            else:
                data += bytes([pad_len] * pad_len)

        return data

    def _unpadData(self, data, pad, padmode):
        # Unpad data depending on the mode.
        if not data:
            return data
        if pad and padmode == PAD_PKCS5:
            raise ValueError("Cannot use a pad character with PAD_PKCS5")
        if padmode is None:
            # Get the default padding mode.
            padmode = self.getPadMode()

        if padmode == PAD_NORMAL:
            if not pad:
                # Get the default padding.
                pad = self.getPadding()
            if pad:
                data = data[: -self.block_size] + data[-self.block_size :].rstrip(pad)

        elif padmode == PAD_PKCS5:
            if _pythonMajorVersion < 3:
                pad_len = ord(data[-1])
            else:
                pad_len = data[-1]
            data = data[:-pad_len]

        return data

    def _guardAgainstUnicode(self, data):
        # Only accept byte strings or ascii unicode values, otherwise
        # there is no way to correctly decode the data into bytes.
        if _pythonMajorVersion < 3:
            if isinstance(data, unicode):
                raise ValueError("pyDes can only work with bytes, not Unicode strings.")
        else:
            if isinstance(data, str):
                # Only accept ascii unicode values.
                try:
                    return data.encode("ascii")
                except UnicodeEncodeError:
                    pass
                raise ValueError(
                    "pyDes can only work with encoded strings, not Unicode."
                )
        return data


#############################################################################
#                   DES                     #
#############################################################################
class des(_baseDes):
    """DES encryption/decrytpion class

    Supports ECB (Electronic Code Book) and CBC (Cypher Block Chaining) modes.

    pyDes.des(key,[mode], [IV])

    key  -> Bytes containing the encryption key, must be exactly 8 bytes
    mode -> Optional argument for encryption type, can be either pyDes.ECB
        (Electronic Code Book), pyDes.CBC (Cypher Block Chaining)
    IV   -> Optional Initial Value bytes, must be supplied if using CBC mode.
        Must be 8 bytes in length.
    pad  -> Optional argument, set the pad character (PAD_NORMAL) to use
        during all encrypt/decrpt operations done with this instance.
    padmode -> Optional argument, set the padding mode (PAD_NORMAL or
        PAD_PKCS5) to use during all encrypt/decrpt operations done
        with this instance.
    """

    # Permutation and translation tables for DES
    __pc1 = [
        56,
        48,
        40,
        32,
        24,
        16,
        8,
        0,
        57,
        49,
        41,
        33,
        25,
        17,
        9,
        1,
        58,
        50,
        42,
        34,
        26,
        18,
        10,
        2,
        59,
        51,
        43,
        35,
        62,
        54,
        46,
        38,
        30,
        22,
        14,
        6,
        61,
        53,
        45,
        37,
        29,
        21,
        13,
        5,
        60,
        52,
        44,
        36,
        28,
        20,
        12,
        4,
        27,
        19,
        11,
        3,
    ]

    # number left rotations of pc1
    __left_rotations = [1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1]

    # permuted choice key (table 2)
    __pc2 = [
        13,
        16,
        10,
        23,
        0,
        4,
        2,
        27,
        14,
        5,
        20,
        9,
        22,
        18,
        11,
        3,
        25,
        7,
        15,
        6,
        26,
        19,
        12,
        1,
        40,
        51,
        30,
        36,
        46,
        54,
        29,
        39,
        50,
        44,
        32,
        47,
        43,
        48,
        38,
        55,
        33,
        52,
        45,
        41,
        49,
        35,
        28,
        31,
    ]

    # initial permutation IP
    __ip = [
        57,
        49,
        41,
        33,
        25,
        17,
        9,
        1,
        59,
        51,
        43,
        35,
        27,
        19,
        11,
        3,
        61,
        53,
        45,
        37,
        29,
        21,
        13,
        5,
        63,
        55,
        47,
        39,
        31,
        23,
        15,
        7,
        56,
        48,
        40,
        32,
        24,
        16,
        8,
        0,
        58,
        50,
        42,
        34,
        26,
        18,
        10,
        2,
        60,
        52,
        44,
        36,
        28,
        20,
        12,
        4,
        62,
        54,
        46,
        38,
        30,
        22,
        14,
        6,
    ]

    # Expansion table for turning 32 bit blocks into 48 bits
    __expansion_table = [
        31,
        0,
        1,
        2,
        3,
        4,
        3,
        4,
        5,
        6,
        7,
        8,
        7,
        8,
        9,
        10,
        11,
        12,
        11,
        12,
        13,
        14,
        15,
        16,
        15,
        16,
        17,
        18,
        19,
        20,
        19,
        20,
        21,
        22,
        23,
        24,
        23,
        24,
        25,
        26,
        27,
        28,
        27,
        28,
        29,
        30,
        31,
        0,
    ]

    # The (in)famous S-boxes
    __sbox = [
        # S1
        [
            14,
            4,
            13,
            1,
            2,
            15,
            11,
            8,
            3,
            10,
            6,
            12,
            5,
            9,
            0,
            7,
            0,
            15,
            7,
            4,
            14,
            2,
            13,
            1,
            10,
            6,
            12,
            11,
            9,
            5,
            3,
            8,
            4,
            1,
            14,
            8,
            13,
            6,
            2,
            11,
            15,
            12,
            9,
            7,
            3,
            10,
            5,
            0,
            15,
            12,
            8,
            2,
            4,
            9,
            1,
            7,
            5,
            11,
            3,
            14,
            10,
            0,
            6,
            13,
        ],
        # S2
        [
            15,
            1,
            8,
            14,
            6,
            11,
            3,
            4,
            9,
            7,
            2,
            13,
            12,
            0,
            5,
            10,
            3,
            13,
            4,
            7,
            15,
            2,
            8,
            14,
            12,
            0,
            1,
            10,
            6,
            9,
            11,
            5,
            0,
            14,
            7,
            11,
            10,
            4,
            13,
            1,
            5,
            8,
            12,
            6,
            9,
            3,
            2,
            15,
            13,
            8,
            10,
            1,
            3,
            15,
            4,
            2,
            11,
            6,
            7,
            12,
            0,
            5,
            14,
            9,
        ],
        # S3
        [
            10,
            0,
            9,
            14,
            6,
            3,
            15,
            5,
            1,
            13,
            12,
            7,
            11,
            4,
            2,
            8,
            13,
            7,
            0,
            9,
            3,
            4,
            6,
            10,
            2,
            8,
            5,
            14,
            12,
            11,
            15,
            1,
            13,
            6,
            4,
            9,
            8,
            15,
            3,
            0,
            11,
            1,
            2,
            12,
            5,
            10,
            14,
            7,
            1,
            10,
            13,
            0,
            6,
            9,
            8,
            7,
            4,
            15,
            14,
            3,
            11,
            5,
            2,
            12,
        ],
        # S4
        [
            7,
            13,
            14,
            3,
            0,
            6,
            9,
            10,
            1,
            2,
            8,
            5,
            11,
            12,
            4,
            15,
            13,
            8,
            11,
            5,
            6,
            15,
            0,
            3,
            4,
            7,
            2,
            12,
            1,
            10,
            14,
            9,
            10,
            6,
            9,
            0,
            12,
            11,
            7,
            13,
            15,
            1,
            3,
            14,
            5,
            2,
            8,
            4,
            3,
            15,
            0,
            6,
            10,
            1,
            13,
            8,
            9,
            4,
            5,
            11,
            12,
            7,
            2,
            14,
        ],
        # S5
        [
            2,
            12,
            4,
            1,
            7,
            10,
            11,
            6,
            8,
            5,
            3,
            15,
            13,
            0,
            14,
            9,
            14,
            11,
            2,
            12,
            4,
            7,
            13,
            1,
            5,
            0,
            15,
            10,
            3,
            9,
            8,
            6,
            4,
            2,
            1,
            11,
            10,
            13,
            7,
            8,
            15,
            9,
            12,
            5,
            6,
            3,
            0,
            14,
            11,
            8,
            12,
            7,
            1,
            14,
            2,
            13,
            6,
            15,
            0,
            9,
            10,
            4,
            5,
            3,
        ],
        # S6
        [
            12,
            1,
            10,
            15,
            9,
            2,
            6,
            8,
            0,
            13,
            3,
            4,
            14,
            7,
            5,
            11,
            10,
            15,
            4,
            2,
            7,
            12,
            9,
            5,
            6,
            1,
            13,
            14,
            0,
            11,
            3,
            8,
            9,
            14,
            15,
            5,
            2,
            8,
            12,
            3,
            7,
            0,
            4,
            10,
            1,
            13,
            11,
            6,
            4,
            3,
            2,
            12,
            9,
            5,
            15,
            10,
            11,
            14,
            1,
            7,
            6,
            0,
            8,
            13,
        ],
        # S7
        [
            4,
            11,
            2,
            14,
            15,
            0,
            8,
            13,
            3,
            12,
            9,
            7,
            5,
            10,
            6,
            1,
            13,
            0,
            11,
            7,
            4,
            9,
            1,
            10,
            14,
            3,
            5,
            12,
            2,
            15,
            8,
            6,
            1,
            4,
            11,
            13,
            12,
            3,
            7,
            14,
            10,
            15,
            6,
            8,
            0,
            5,
            9,
            2,
            6,
            11,
            13,
            8,
            1,
            4,
            10,
            7,
            9,
            5,
            0,
            15,
            14,
            2,
            3,
            12,
        ],
        # S8
        [
            13,
            2,
            8,
            4,
            6,
            15,
            11,
            1,
            10,
            9,
            3,
            14,
            5,
            0,
            12,
            7,
            1,
            15,
            13,
            8,
            10,
            3,
            7,
            4,
            12,
            5,
            6,
            11,
            0,
            14,
            9,
            2,
            7,
            11,
            4,
            1,
            9,
            12,
            14,
            2,
            0,
            6,
            10,
            13,
            15,
            3,
            5,
            8,
            2,
            1,
            14,
            7,
            4,
            10,
            8,
            13,
            15,
            12,
            9,
            0,
            3,
            5,
            6,
            11,
        ],
    ]

    # 32-bit permutation function P used on the output of the S-boxes
    __p = [
        15,
        6,
        19,
        20,
        28,
        11,
        27,
        16,
        0,
        14,
        22,
        25,
        4,
        17,
        30,
        9,
        1,
        7,
        23,
        13,
        31,
        26,
        2,
        8,
        18,
        12,
        29,
        5,
        21,
        10,
        3,
        24,
    ]

    # final permutation IP^-1
    __fp = [
        39,
        7,
        47,
        15,
        55,
        23,
        63,
        31,
        38,
        6,
        46,
        14,
        54,
        22,
        62,
        30,
        37,
        5,
        45,
        13,
        53,
        21,
        61,
        29,
        36,
        4,
        44,
        12,
        52,
        20,
        60,
        28,
        35,
        3,
        43,
        11,
        51,
        19,
        59,
        27,
        34,
        2,
        42,
        10,
        50,
        18,
        58,
        26,
        33,
        1,
        41,
        9,
        49,
        17,
        57,
        25,
        32,
        0,
        40,
        8,
        48,
        16,
        56,
        24,
    ]

    # Type of crypting being done
    ENCRYPT = 0x00
    DECRYPT = 0x01

    # Initialisation
    def __init__(self, key, mode=ECB, IV=None, pad=None, padmode=PAD_NORMAL):
        # Sanity checking of arguments.
        if len(key) != 8:
            raise ValueError("Invalid DES key size. Key must be exactly 8 bytes long.")
        _baseDes.__init__(self, mode, IV, pad, padmode)
        self.key_size = 8

        self.L = []
        self.R = []
        self.Kn = [[0] * 48] * 16  # 16 48-bit keys (K1 - K16)
        self.final = []

        self.setKey(key)

    def setKey(self, key):
        """Will set the crypting key for this object. Must be 8 bytes."""
        _baseDes.setKey(self, key)
        self.__create_sub_keys()

    def __String_to_BitList(self, data):
        """Turn the string data, into a list of bits (1, 0)'s"""
        if _pythonMajorVersion < 3:
            # Turn the strings into integers. Python 3 uses a bytes
            # class, which already has this behaviour.
            data = [ord(c) for c in data]
        l = len(data) * 8
        result = [0] * l
        pos = 0
        for ch in data:
            i = 7
            while i >= 0:
                if ch & (1 << i) != 0:
                    result[pos] = 1
                else:
                    result[pos] = 0
                pos += 1
                i -= 1

        return result

    def __BitList_to_String(self, data):
        """Turn the list of bits -> data, into a string"""
        result = []
        pos = 0
        c = 0
        while pos < len(data):
            c += data[pos] << (7 - (pos % 8))
            if (pos % 8) == 7:
                result.append(c)
                c = 0
            pos += 1

        if _pythonMajorVersion < 3:
            return "".join([chr(c) for c in result])
        else:
            return bytes(result)

    def __permutate(self, table, block):
        """Permutate this block with the specified table"""
        return list(map(lambda x: block[x], table))

    # Transform the secret key, so that it is ready for data processing
    # Create the 16 subkeys, K[1] - K[16]
    def __create_sub_keys(self):
        """Create the 16 subkeys K[1] to K[16] from the given key"""
        key = self.__permutate(des.__pc1, self.__String_to_BitList(self.getKey()))
        i = 0
        # Split into Left and Right sections
        self.L = key[:28]
        self.R = key[28:]
        while i < 16:
            j = 0
            # Perform circular left shifts
            while j < des.__left_rotations[i]:
                self.L.append(self.L[0])
                del self.L[0]

                self.R.append(self.R[0])
                del self.R[0]

                j += 1

            # Create one of the 16 subkeys through pc2 permutation
            self.Kn[i] = self.__permutate(des.__pc2, self.L + self.R)

            i += 1

    # Main part of the encryption algorithm, the number cruncher :)
    def __des_crypt(self, block, crypt_type):
        """Crypt the block of data through DES bit-manipulation"""
        block = self.__permutate(des.__ip, block)
        self.L = block[:32]
        self.R = block[32:]

        # Encryption starts from Kn[1] through to Kn[16]
        if crypt_type == des.ENCRYPT:
            iteration = 0
            iteration_adjustment = 1
        # Decryption starts from Kn[16] down to Kn[1]
        else:
            iteration = 15
            iteration_adjustment = -1

        i = 0
        while i < 16:
            # Make a copy of R[i-1], this will later become L[i]
            tempR = self.R[:]

            # Permutate R[i - 1] to start creating R[i]
            self.R = self.__permutate(des.__expansion_table, self.R)

            # Exclusive or R[i - 1] with K[i], create B[1] to B[8] whilst here
            self.R = list(map(lambda x, y: x ^ y, self.R, self.Kn[iteration]))
            B = [
                self.R[:6],
                self.R[6:12],
                self.R[12:18],
                self.R[18:24],
                self.R[24:30],
                self.R[30:36],
                self.R[36:42],
                self.R[42:],
            ]
            # Optimization: Replaced below commented code with above
            # j = 0
            # B = []
            # while j < len(self.R):
            #   self.R[j] = self.R[j] ^ self.Kn[iteration][j]
            #   j += 1
            #   if j % 6 == 0:
            #       B.append(self.R[j-6:j])

            # Permutate B[1] to B[8] using the S-Boxes
            j = 0
            Bn = [0] * 32
            pos = 0
            while j < 8:
                # Work out the offsets
                m = (B[j][0] << 1) + B[j][5]
                n = (B[j][1] << 3) + (B[j][2] << 2) + (B[j][3] << 1) + B[j][4]

                # Find the permutation value
                v = des.__sbox[j][(m << 4) + n]

                # Turn value into bits, add it to result: Bn
                Bn[pos] = (v & 8) >> 3
                Bn[pos + 1] = (v & 4) >> 2
                Bn[pos + 2] = (v & 2) >> 1
                Bn[pos + 3] = v & 1

                pos += 4
                j += 1

            # Permutate the concatination of B[1] to B[8] (Bn)
            self.R = self.__permutate(des.__p, Bn)

            # Xor with L[i - 1]
            self.R = list(map(lambda x, y: x ^ y, self.R, self.L))
            # Optimization: This now replaces the below commented code
            # j = 0
            # while j < len(self.R):
            #   self.R[j] = self.R[j] ^ self.L[j]
            #   j += 1

            # L[i] becomes R[i - 1]
            self.L = tempR

            i += 1
            iteration += iteration_adjustment

        # Final permutation of R[16]L[16]
        self.final = self.__permutate(des.__fp, self.R + self.L)
        return self.final

    # Data to be encrypted/decrypted
    def crypt(self, data, crypt_type):
        """Crypt the data in blocks, running it through des_crypt()"""

        # Error check the data
        if not data:
            return ""
        if len(data) % self.block_size != 0:
            if crypt_type == des.DECRYPT:  # Decryption must work on 8 byte blocks
                raise ValueError(
                    "Invalid data length, data must be a multiple of "
                    + str(self.block_size)
                    + " bytes\n."
                )
            if not self.getPadding():
                raise ValueError(
                    "Invalid data length, data must be a multiple of "
                    + str(self.block_size)
                    + " bytes\n. Try setting the optional padding character"
                )
            else:
                data += (
                    self.block_size - (len(data) % self.block_size)
                ) * self.getPadding()
            # print "Len of data: %f" % (len(data) / self.block_size)

        if self.getMode() == CBC:
            if self.getIV():
                iv = self.__String_to_BitList(self.getIV())
            else:
                raise ValueError(
                    "For CBC mode, you must supply the Initial Value (IV) for ciphering"
                )

        # Split the data into blocks, crypting each one separately
        i = 0
        dict = {}
        result = []
        # cached = 0
        # lines = 0
        while i < len(data):
            # Test code for caching encryption results
            # lines += 1
            # if dict.has_key(data[i:i+8]):
            # print "Cached result for: %s" % data[i:i+8]
            #   cached += 1
            #   result.append(dict[data[i:i+8]])
            #   i += 8
            #   continue

            block = self.__String_to_BitList(data[i : i + 8])

            # Xor with IV if using CBC mode
            if self.getMode() == CBC:
                if crypt_type == des.ENCRYPT:
                    block = list(map(lambda x, y: x ^ y, block, iv))
                    # j = 0
                    # while j < len(block):
                    #   block[j] = block[j] ^ iv[j]
                    #   j += 1

                processed_block = self.__des_crypt(block, crypt_type)

                if crypt_type == des.DECRYPT:
                    processed_block = list(map(lambda x, y: x ^ y, processed_block, iv))
                    # j = 0
                    # while j < len(processed_block):
                    #   processed_block[j] = processed_block[j] ^ iv[j]
                    #   j += 1
                    iv = block
                else:
                    iv = processed_block
            else:
                processed_block = self.__des_crypt(block, crypt_type)

            # Add the resulting crypted block to our list
            # d = self.__BitList_to_String(processed_block)
            # result.append(d)
            result.append(self.__BitList_to_String(processed_block))
            # dict[data[i:i+8]] = d
            i += 8

        # print "Lines: %d, cached: %d" % (lines, cached)

        # Return the full crypted string
        if _pythonMajorVersion < 3:
            return "".join(result)
        else:
            return bytes.fromhex("").join(result)

    def encrypt(self, data, pad=None, padmode=None):
        """encrypt(data, [pad], [padmode]) -> bytes

        data : Bytes to be encrypted
        pad  : Optional argument for encryption padding. Must only be one byte
        padmode : Optional argument for overriding the padding mode.

        The data must be a multiple of 8 bytes and will be encrypted
        with the already specified key. Data does not have to be a
        multiple of 8 bytes if the padding character is supplied, or
        the padmode is set to PAD_PKCS5, as bytes will then added to
        ensure the be padded data is a multiple of 8 bytes.
        """
        data = self._guardAgainstUnicode(data)
        if pad is not None:
            pad = self._guardAgainstUnicode(pad)
        data = self._padData(data, pad, padmode)
        return self.crypt(data, des.ENCRYPT)

    def decrypt(self, data, pad=None, padmode=None):
        """decrypt(data, [pad], [padmode]) -> bytes

        data : Bytes to be encrypted
        pad  : Optional argument for decryption padding. Must only be one byte
        padmode : Optional argument for overriding the padding mode.

        The data must be a multiple of 8 bytes and will be decrypted
        with the already specified key. In PAD_NORMAL mode, if the
        optional padding character is supplied, then the un-encrypted
        data will have the padding characters removed from the end of
        the bytes. This pad removal only occurs on the last 8 bytes of
        the data (last data block). In PAD_PKCS5 mode, the special
        padding end markers will be removed from the data after decrypting.
        """
        data = self._guardAgainstUnicode(data)
        if pad is not None:
            pad = self._guardAgainstUnicode(pad)
        data = self.crypt(data, des.DECRYPT)
        return self._unpadData(data, pad, padmode)


#############################################################################
#               Triple DES                  #
#############################################################################
class triple_des(_baseDes):
    """Triple DES encryption/decrytpion class

    This algorithm uses the DES-EDE3 (when a 24 byte key is supplied) or
    the DES-EDE2 (when a 16 byte key is supplied) encryption methods.
    Supports ECB (Electronic Code Book) and CBC (Cypher Block Chaining) modes.

    pyDes.des(key, [mode], [IV])

    key  -> Bytes containing the encryption key, must be either 16 or
            24 bytes long
    mode -> Optional argument for encryption type, can be either pyDes.ECB
        (Electronic Code Book), pyDes.CBC (Cypher Block Chaining)
    IV   -> Optional Initial Value bytes, must be supplied if using CBC mode.
        Must be 8 bytes in length.
    pad  -> Optional argument, set the pad character (PAD_NORMAL) to use
        during all encrypt/decrpt operations done with this instance.
    padmode -> Optional argument, set the padding mode (PAD_NORMAL or
        PAD_PKCS5) to use during all encrypt/decrpt operations done
        with this instance.
    """

    def __init__(self, key, mode=ECB, IV=None, pad=None, padmode=PAD_NORMAL):
        _baseDes.__init__(self, mode, IV, pad, padmode)
        self.setKey(key)

    def setKey(self, key):
        """Will set the crypting key for this object. Either 16 or 24 bytes long."""
        self.key_size = 24  # Use DES-EDE3 mode
        if len(key) != self.key_size:
            if len(key) == 16:  # Use DES-EDE2 mode
                self.key_size = 16
            else:
                raise ValueError(
                    "Invalid triple DES key size. Key must be either 16 or 24 bytes long"
                )
        if self.getMode() == CBC:
            if not self.getIV():
                # Use the first 8 bytes of the key
                self._iv = key[: self.block_size]
            if len(self.getIV()) != self.block_size:
                raise ValueError("Invalid IV, must be 8 bytes in length")
        self.__key1 = des(key[:8], self._mode, self._iv, self._padding, self._padmode)
        self.__key2 = des(key[8:16], self._mode, self._iv, self._padding, self._padmode)
        if self.key_size == 16:
            self.__key3 = self.__key1
        else:
            self.__key3 = des(
                key[16:], self._mode, self._iv, self._padding, self._padmode
            )
        _baseDes.setKey(self, key)

    # Override setter methods to work on all 3 keys.

    def setMode(self, mode):
        """Sets the type of crypting mode, pyDes.ECB or pyDes.CBC"""
        _baseDes.setMode(self, mode)
        for key in (self.__key1, self.__key2, self.__key3):
            key.setMode(mode)

    def setPadding(self, pad):
        """setPadding() -> bytes of length 1. Padding character."""
        _baseDes.setPadding(self, pad)
        for key in (self.__key1, self.__key2, self.__key3):
            key.setPadding(pad)

    def setPadMode(self, mode):
        """Sets the type of padding mode, pyDes.PAD_NORMAL or pyDes.PAD_PKCS5"""
        _baseDes.setPadMode(self, mode)
        for key in (self.__key1, self.__key2, self.__key3):
            key.setPadMode(mode)

    def setIV(self, IV):
        """Will set the Initial Value, used in conjunction with CBC mode"""
        _baseDes.setIV(self, IV)
        for key in (self.__key1, self.__key2, self.__key3):
            key.setIV(IV)

    def encrypt(self, data, pad=None, padmode=None):
        """encrypt(data, [pad], [padmode]) -> bytes

        data : bytes to be encrypted
        pad  : Optional argument for encryption padding. Must only be one byte
        padmode : Optional argument for overriding the padding mode.

        The data must be a multiple of 8 bytes and will be encrypted
        with the already specified key. Data does not have to be a
        multiple of 8 bytes if the padding character is supplied, or
        the padmode is set to PAD_PKCS5, as bytes will then added to
        ensure the be padded data is a multiple of 8 bytes.
        """
        ENCRYPT = des.ENCRYPT
        DECRYPT = des.DECRYPT
        data = self._guardAgainstUnicode(data)
        if pad is not None:
            pad = self._guardAgainstUnicode(pad)
        # Pad the data accordingly.
        data = self._padData(data, pad, padmode)
        if self.getMode() == CBC:
            self.__key1.setIV(self.getIV())
            self.__key2.setIV(self.getIV())
            self.__key3.setIV(self.getIV())
            i = 0
            result = []
            while i < len(data):
                block = self.__key1.crypt(data[i : i + 8], ENCRYPT)
                block = self.__key2.crypt(block, DECRYPT)
                block = self.__key3.crypt(block, ENCRYPT)
                self.__key1.setIV(block)
                self.__key2.setIV(block)
                self.__key3.setIV(block)
                result.append(block)
                i += 8
            if _pythonMajorVersion < 3:
                return "".join(result)
            else:
                return bytes.fromhex("").join(result)
        else:
            data = self.__key1.crypt(data, ENCRYPT)
            data = self.__key2.crypt(data, DECRYPT)
            return self.__key3.crypt(data, ENCRYPT)

    def decrypt(self, data, pad=None, padmode=None):
        """decrypt(data, [pad], [padmode]) -> bytes

        data : bytes to be encrypted
        pad  : Optional argument for decryption padding. Must only be one byte
        padmode : Optional argument for overriding the padding mode.

        The data must be a multiple of 8 bytes and will be decrypted
        with the already specified key. In PAD_NORMAL mode, if the
        optional padding character is supplied, then the un-encrypted
        data will have the padding characters removed from the end of
        the bytes. This pad removal only occurs on the last 8 bytes of
        the data (last data block). In PAD_PKCS5 mode, the special
        padding end markers will be removed from the data after
        decrypting, no pad character is required for PAD_PKCS5.
        """
        ENCRYPT = des.ENCRYPT
        DECRYPT = des.DECRYPT
        data = self._guardAgainstUnicode(data)
        if pad is not None:
            pad = self._guardAgainstUnicode(pad)
        if self.getMode() == CBC:
            self.__key1.setIV(self.getIV())
            self.__key2.setIV(self.getIV())
            self.__key3.setIV(self.getIV())
            i = 0
            result = []
            while i < len(data):
                iv = data[i : i + 8]
                block = self.__key3.crypt(iv, DECRYPT)
                block = self.__key2.crypt(block, ENCRYPT)
                block = self.__key1.crypt(block, DECRYPT)
                self.__key1.setIV(iv)
                self.__key2.setIV(iv)
                self.__key3.setIV(iv)
                result.append(block)
                i += 8
            if _pythonMajorVersion < 3:
                data = "".join(result)
            else:
                data = bytes.fromhex("").join(result)
        else:
            data = self.__key3.crypt(data, DECRYPT)
            data = self.__key2.crypt(data, ENCRYPT)
            data = self.__key1.crypt(data, DECRYPT)
        return self._unpadData(data, pad, padmode)



import os
from datetime import datetime, timedelta
from pathlib import Path

from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from twisted.application import internet
from twisted.web.resource import EncodingResourceWrapper
from twisted.web.server import Site, GzipEncoderFactory
from twisted.internet.ssl import DefaultOpenSSLContextFactory

from opencanary.modules import CanaryService
from opencanary.modules.http import (
    BasicLogin,
    CanaryHTTP,
    RedirectCustomHeaders,
    StaticNoDirListing,
)


class CanaryHTTPS(CanaryService):
    NAME = "https"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.skin = config.getVal("https.skin", default="basicLogin")
        # We share the skin dir with HTTP rather than duplicate everything
        self.skindir = config.getVal("http.skindir", default="")
        if not os.path.isdir(self.skindir):
            self.skindir = os.path.join(CanaryHTTP.resource_dir(), "skin", self.skin)
        self.staticdir = os.path.join(self.skindir, "static")
        self.port = int(config.getVal("https.port", default=443))
        ubanner = config.getVal("http.banner", default="Apache/2.2.22 (Ubuntu)")
        self.banner = ubanner.encode("utf8")
        StaticNoDirListing.BANNER = self.banner
        self.listen_addr = config.getVal("device.listen_addr", default="")
        self.domain_name = config.getVal(
            "https.domain_name", default="synologynas.local"
        )
        self.certificate_path = Path(
            config.getVal(
                "https.certificate", default="/etc/ssl/opencanary/opencanary.pem"
            )
        )
        self.key_path = Path(
            config.getVal("https.key", default="/etc/ssl/opencanary/opencanary.key")
        )
        self.load_certificates()

    def load_certificates(self):
        """
        If certificates already exist, use them, otherwise we generate self signed certificates.
        """
        if self.certificate_path.exists() and self.key_path.exists():
            pass
        else:
            # Make the directory to save the keys into (if it doesn't already exist)
            self.key_path.parent.mkdir(parents=True, exist_ok=True)
            self.certificate_path.parent.mkdir(parents=True, exist_ok=True)

            # Generate our Key
            key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=2048,
            )
            # Write our key to disk for safe keeping
            with open(self.key_path, "wb") as key_file:
                key_file.write(
                    key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.NoEncryption(),
                    )
                )
            # Various details about who we are. For a self-signed certificate the
            # subject and issuer are always the same.
            subject = issuer = x509.Name(
                [
                    x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
                    x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
                    x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
                    x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Synology Inc. CA"),
                    x509.NameAttribute(NameOID.COMMON_NAME, self.domain_name),
                ]
            )
            cert = (
                x509.CertificateBuilder()
                .subject_name(subject)
                .issuer_name(issuer)
                .public_key(key.public_key())
                .serial_number(x509.random_serial_number())
                .not_valid_before(datetime.utcnow())
                .not_valid_after(datetime.utcnow() + timedelta(days=365))
                .sign(key, hashes.SHA256())
            )
            # Write our certificate out to disk.
            with open(self.certificate_path, "wb") as f:
                f.write(cert.public_bytes(serialization.Encoding.PEM))

    def getService(self):
        page = BasicLogin(factory=self)
        root = StaticNoDirListing(self.staticdir)
        root.createErrorPages(self)
        root.putChild(b"", RedirectCustomHeaders(b"/index.html", factory=self))
        root.putChild(b"index.html", page)
        wrapped = EncodingResourceWrapper(root, [GzipEncoderFactory()])
        site = Site(wrapped)
        return internet.SSLServer(
            self.port,
            site,
            DefaultOpenSSLContextFactory(
                privateKeyFileName=self.key_path,
                certificateFileName=self.certificate_path,
            ),
            interface=self.listen_addr,
        )



from opencanary.modules import CanaryService
from opencanary.config import ConfigException

from twisted.protocols.policies import TimeoutMixin
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet
from ntlmlib.messages import ChallengeResponse, TargetInfo

import struct
import collections

# Monkeypatch bug in ntmllib
if getattr(TargetInfo, "getData", None) is None:

    def getData(self):
        return self.get_data()

    TargetInfo.getData = getData

TDSPacket = collections.namedtuple(
    "TDSPacket", "type status spid packetid window payload"
)
PreLoginOption = collections.namedtuple("PreLoginOption", "token data")


class MSSQLProtocol(Protocol, TimeoutMixin):
    # overview https://msdn.microsoft.com/en-us/library/dd357422.aspx

    TDS_HEADER_LEN = 8
    TDS_TYPE_PRELOGIN = 0x12
    TDS_TYPE_RESPONSE = 0x04
    TDS_TYPE_LOGIN7 = 0x10
    TDS_TYPE_SSPI = 0x11

    # https://msdn.microsoft.com/en-us/library/dd357559.aspx
    PRELOGIN_VERSION = 0x00
    PRELOGIN_ENCRYPTION = 0x01
    PRELOGIN_INSTOPT = 0x02
    PRELOGIN_THREADID = 0x03
    PRELOGIN_MARS = 0x04
    PRELOGIN_TRACEID = 0x05
    PRELOGIN_FEDAUTHREQUIRED = 0x06
    PRELOGIN_NONCEOPT = 0x07

    # https://msdn.microsoft.com/en-us/library/dd304019.aspx
    LOGIN7_OPT1_BYTEORDER_MASK = 0x01
    LOGIN7_OPT1_BYTEORDER_X86 = 0x01
    LOGIN7_OPT1_BYTEORDER_68000 = 0x01
    LOGIN7_FIELDS = [
        "length",
        "TDSversion",
        "packetSize",
        "clientProgVer",
        "clientPID",
        "connID",
        "optFlags1",
        "optFlags2",
        "typeFlags",
        "optFlags3",
        "clientTimeZone",
        "clientLCID",
        "ibHostName",
        "cchHostName",
        "ibUserName",
        "cchUserName",
        "ibPassword",
        "cchPassword",
        "ibAppName",
        "cchAppName",
        "ibServerName",
        "cchServerName",
        "ibExtension",
        "cbExtension",
        "ibCltIntName",
        "cchCltIntName",
        "ibLanguage",
        "cchLanguage",
        "ibDatabase",
        "cchDatabase",
        "ClientID",
        "ibSSPI",
        "cbSSPI",
        "ibAtchDBFile",
        "cchAtchDBFile",
        "ibChangePassword",
        "cchChangePassword",
        "cbSSPILong",
    ]

    NMAP_PROBE_1 = TDSPacket(
        type=18,
        status=1,
        spid=0,
        packetid=0,
        window=0,
        payload=b"\x00\x00\x15\x00\x06\x01\x00\x1b\x00\x01\x02\x00\x1c\x00\x0c\x03\x00(\x00\x04\xff\x08\x00\x01U\x00\x00\x00MSSQLServer\x00H\x0f\x00\x00",
    )

    NMAP_PROBE_1_RESP = {
        "2008R2": b"\x04\x01\x00.\x00\x00\x01\x00\x00\x00\x15\x00\x06\x01\x00\x1b\x00\x01\x02\x00\x1c\x00\x01\x03\x00\x1d\x00\x00\xff\x0a\x32\x10\xb4",
        "2012": b"\x04\x01\x00\x25\x00\x00\x01\x00\x00\x00\x15\x00\x06\x01\x00\x1b\x00\x01\x02\x00\x1c\x00\x01\x03\x00\x1d\x00\x00\xff\x0b\x00\x0c\x38",
        "2014": b"\x04\x01\x00\x25\x00\x00\x01\x00\x00\x00\x15\x00\x06\x01\x00\x1b\x00\x01\x02\x00\x1c\x00\x01\x03\x00\x1d\x00\x00\xff\x0c\x00\x07\xd0",
    }

    def __init__(self, factory):
        self._busyReceiving = False
        self._buffer = b""
        self.factory = factory
        self.setTimeout(10)

    @staticmethod
    def build_packet(tds):
        header = struct.pack(
            ">BBHHBB",
            tds.type,
            tds.status,
            len(tds.payload) + MSSQLProtocol.TDS_HEADER_LEN,
            tds.spid,
            tds.packetid,
            tds.window,
        )

        return header + tds.payload

    @staticmethod
    def parsePreLogin(data):
        # parse the initial packet
        index = data.find("\xff")

        if index <= 0:
            return None

        options = data[:index]

        if len(options) % 5 != 0:
            return None

        def getOption(i):
            (token, offset, length) = struct.unpack(">BHH", options[i : i + 5])
            tokendata = data[offset : offset + length]
            assert len(tokendata) == length
            return PreLoginOption._make((token, tokendata))

        try:
            return map(getOption, range(0, len(options), 5))
        except Exception as e:
            print(e)

        return None

    @staticmethod
    def buildPreLogin(preloginopts):
        preloginopts.sort(key=lambda x: x.token)
        dataoffset = len(preloginopts) * 5 + len("\xff")
        data = b""
        header = b""

        for opt in preloginopts:
            token = opt.token
            offset = dataoffset + len(data)
            length = len(opt.data)
            header += struct.pack(">BHH", token, offset, length)
            data += opt.data

        return header + b"\xff" + data

    @staticmethod
    def parseLogin7(data):
        # These header values apper to be little endian, even though
        # the byte-order flag suggests otherwise. In case it is
        # dependent, simply check the endian flag in optFlags1 in
        # before parsing the rest in ord(data[24]). For the moment,
        # we'll assume that flag applies only to the data part.
        hfmt = "< 6I 4B l I 18H 6s 6H I"
        hlen = struct.calcsize(hfmt)
        try:
            htuple = struct.unpack(hfmt, data[:hlen])
        except Exception:
            return None

        def decodePassChar(c):
            # https://msdn.microsoft.com/en-us/library/dd304019.aspx
            c = ord(c) ^ 0xA5
            return chr(((c & 0x0F) << 4) | (c >> 4))

        fields = {}
        for (i, fieldname) in enumerate(MSSQLProtocol.LOGIN7_FIELDS):
            fields[fieldname] = htuple[i]

        loginData = {}
        for (
            field
        ) in "HostName UserName Password AppName ServerName Language Database CltIntName".split():
            try:
                findex = fields["ib" + field]
                flen = (
                    fields["cch" + field] * 2
                )  # this is character count, not count of bytes
                _fdata = data[findex : findex + flen]
                if field == "Password":
                    _fdata = "".join(map(decodePassChar, _fdata))
                loginData[field] = _fdata.decode("utf-16")
            except Exception:
                pass

        field = "SSPI"
        findex = fields["ib" + field]
        flen = fields["cb" + field]
        if flen == 0:
            return loginData
        loginData["NTLM"] = data[findex : findex + flen]
        return loginData

    @staticmethod
    def buildError(msgtxt, serverName, procName=""):
        tokentype = 0xAA
        number = 18456
        state = 1
        sevclass = 14
        msgtxtlen = len(msgtxt)
        serverNameLen = len(serverName)
        procNameLen = len(procName)
        lineNumber = 1

        msgtxt = msgtxt.encode("utf-16le")
        procName = procName.encode("utf-16le")
        serverName = serverName.encode("utf-16le")

        # This length works, but tests seem show mssql 2014 uses an (incorrect?) length -3 bytes shorter than this
        length = (
            4
            + 1
            + 1
            + 2
            + 1
            + 1
            + 4
            + 2
            + 1
            + len(msgtxt)
            + len(serverName)
            + len(procName)
        )

        fmt = "<BHlBB H%ds B%ds B%ds l" % (len(msgtxt), len(serverName), len(procName))

        data = struct.pack(
            fmt,
            tokentype,
            length,
            number,
            state,
            sevclass,
            msgtxtlen,
            msgtxt,
            serverNameLen,
            serverName,
            procNameLen,
            procName,
            lineNumber,
        )

        return data

    def consume_packet(self):
        """Consume TDS packet off buffer"""
        hlen = MSSQLProtocol.TDS_HEADER_LEN
        if len(self._buffer) < hlen:
            return None

        try:
            header = list(struct.unpack(">BBHHBB", self._buffer[:hlen]))
            plen = header[2]
            del header[2]
            if len(self._buffer) >= plen:
                payload = self._buffer[hlen:plen]
                self._buffer = self._buffer[plen:]
                tds = TDSPacket._make(header + [payload])
                return tds
            else:
                # Whole payload not yet received. Leave header in
                # buffer but return copy of header anyway
                return TDSPacket._make(header + [None])
        except Exception as e:
            print(e)

        return None

    @staticmethod
    def buildChallengeToken():
        spnegoheader = b"\xa1\x82\x01Y0\x82\x01U\xa0\x03\n\x01\x01\xa1\x0c\x06\n+\x06\x01\x04\x01\x827\x02\x02\n\xa2\x82\x01>\x04\x82\x01:"

        # This NTLMSSP challenge is a modified actual challenge
        # stripped using the ntlmlib:
        #
        # # do dynamic patch above to fix ntlmlib bug if nec
        # c = Challenge()
        # c.from_string(raw_str)
        # # c['target_name'] = "yoyoma".encode('utf-16le')
        # del c['target_info'].fields[3]
        # del c['target_info'].fields[4]
        # c['target_info_len'] = ''
        # template = c.get_data()
        old = b"NTLMSSP\x00\x02\x00\x00\x00\x1e\x00\x1e\x008\x00\x00\x00\x15\xc2\x8a\xe26Ph\x8a\xae\x84R\xbe@\xd5qt\xe4\x00\x00\x00\xe4\x00\xe4\x00V\x00\x00\x00\x06\x02\xf0#\x00\x00\x00\x0fW\x00I\x00N\x002\x00K\x001\x002\x00-\x00D\x00O\x00M\x00A\x00I\x00N\x00S\x00\x02\x00\x1e\x00W\x00I\x00N\x002\x00K\x001\x002\x00-\x00D\x00O\x00M\x00A\x00I\x00N\x00S\x00\x01\x00\x1e\x00W\x00I\x00N\x002\x00K\x001\x002\x00-\x00D\x00O\x00M\x00A\x00I\x00N\x00S\x00\x04\x00D\x00w\x00i\x00n\x002\x00k\x001\x002\x00-\x00d\x00o\x00m\x00a\x00i\x00n\x00s\x00r\x00v\x00.\x00c\x00o\x00r\x00p\x00.\x00t\x00h\x00i\x00n\x00k\x00s\x00t\x00.\x00c\x00o\x00m\x00\x03\x00D\x00w\x00i\x00n\x002\x00k\x001\x002\x00-\x00d\x00o\x00m\x00a\x00i\x00n\x00s\x00r\x00v\x00.\x00c\x00o\x00r\x00p\x00.\x00t\x00h\x00i\x00n\x00k\x00s\x00t\x00.\x00c\x00o\x00m\x00\x07\x00\x08\x00\xa2\x9e\xda\x91\x1f\xbb\xd0\x01\x00\x00\x00\x00\x06\x02\xf0#\x00\x00\x00\x0fy\x00o\x00y\x00o\x00m\x00a\x00\x00\x00\x00\x00"

        payload = spnegoheader + old
        return b"\xed" + struct.pack("<H", len(payload)) + payload

    def process(self, tds):
        if tds.payload is None:
            return

        if tds == MSSQLProtocol.NMAP_PROBE_1:
            self.transport.write(
                MSSQLProtocol.NMAP_PROBE_1_RESP[self.factory.canaryservice.version]
            )

        elif tds.type == MSSQLProtocol.TDS_TYPE_PRELOGIN:
            rPreLogin = [
                PreLoginOption(
                    MSSQLProtocol.PRELOGIN_VERSION, b"\x0c\x00\x10\x04\x00\x00"
                ),
                PreLoginOption(MSSQLProtocol.PRELOGIN_ENCRYPTION, b"\x02"),
                PreLoginOption(MSSQLProtocol.PRELOGIN_INSTOPT, b"\x00"),
                PreLoginOption(MSSQLProtocol.PRELOGIN_THREADID, b""),
                PreLoginOption(MSSQLProtocol.PRELOGIN_MARS, b"\x00"),
                PreLoginOption(MSSQLProtocol.PRELOGIN_TRACEID, b""),
            ]

            payload = self.buildPreLogin(rPreLogin)

            rtds = TDSPacket(
                type=MSSQLProtocol.TDS_TYPE_RESPONSE,
                status=0x01,
                spid=0x00,
                packetid=0x01,
                window=0x00,
                payload=payload,
            )
            self.transport.write(self.build_packet(rtds))

        elif tds.type == MSSQLProtocol.TDS_TYPE_LOGIN7:
            loginData = self.parseLogin7(tds.payload)
            if loginData is None:
                self.transport.abortConnection()

            errormsg = ""
            servername = ""
            ntlm = loginData.pop("NTLM", None)
            if ntlm is not None:
                # TODO: handle NTLM challenge generation correctly
                errormsg = "Login failed."
                logdata = {"USERNAME": "", "PASSWORD": ""}
                logtype = self.factory.canaryservice.logger.LOG_MSSQL_LOGIN_WINAUTH
                log = self.factory.canaryservice.log
                log(logdata, transport=self.transport, logtype=logtype)

            else:
                logtype = self.factory.canaryservice.logger.LOG_MSSQL_LOGIN_SQLAUTH
                log = self.factory.canaryservice.log
                log(loginData, transport=self.transport, logtype=logtype)
                username = loginData.get("UserName", "")
                errormsg = "Login failed for user %s." % username

            payload = self.buildError(errormsg, servername)
            # extra data observed on the wire
            payload += b"\xfd\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

            rtds = TDSPacket(
                type=MSSQLProtocol.TDS_TYPE_RESPONSE,
                status=0x01,
                spid=54,
                packetid=0x01,
                window=0x00,
                payload=payload,
            )

            self.transport.write(self.build_packet(rtds))

        elif tds.type == MSSQLProtocol.TDS_TYPE_SSPI:
            # FIXME: parse the SNEGO header correctly to extract the NTLM message
            i = tds.payload.find(b"NTLMSSP\x00")
            if i < 0:
                self.transport.abortConnection()
            ntlmtoken = tds.payload[i:]

            c = ChallengeResponse(0, 0, 0, "unspecified", "unspecified")
            c.from_string(ntlmtoken)
            username = c["user_name"].decode("utf-16le")
            hostname = c["host_name"].decode("utf-16le")
            domain = c["domain_name"].decode("utf-16le")
            loginData = {
                "HOSTNAME": hostname,
                "DOMAINNAME": domain,
            }
            logtype = self.factory.logger.LOG_MSSQL_LOGIN_WINAUTH
            self.logAuth(username, None, loginData, logtype)

            payload = self.buildError(
                "Login failed for user %s\\%s." % (domain, username), hostname
            )
            # extra data observed on the wire
            payload += b"\xfd\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

            rtds = TDSPacket(
                type=MSSQLProtocol.TDS_TYPE_RESPONSE,
                status=0x01,
                spid=54,
                packetid=0x01,
                window=0x00,
                payload=payload,
            )

            self.transport.write(self.build_packet(rtds))

        elif tds.type == 128:
            # initial nmap probe: we're expected to reset connection
            self.transport.abortConnection()
        else:
            self.transport.abortConnection()

    def dataReceived(self, data):
        self._buffer += data
        self.resetTimeout()

        if self._busyReceiving:
            return

        try:
            self._busyReceiving = True
            tds = self.consume_packet()
            if tds is not None:
                self.process(tds)
        finally:
            self._busyReceiving = False

    def timeoutConnection(self):
        self.transport.abortConnection()


class SQLFactory(Factory):
    def __init__(self):
        pass

    def buildProtocol(self, addr):
        return MSSQLProtocol(self)


class MSSQL(CanaryService):
    NAME = "mssql"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("mssql.port", default=1433))
        self.version = config.getVal("mssql.version", default="2012")
        self.listen_addr = config.getVal("device.listen_addr", default="")
        if self.version not in MSSQLProtocol.NMAP_PROBE_1_RESP:
            raise ConfigException("mssql.version", "Invalid MSSQL Version")

    def getService(self):
        factory = SQLFactory()
        factory.canaryservice = self
        return internet.TCPServer(self.port, factory, interface=self.listen_addr)



from opencanary.modules import CanaryService

from zope.interface import implementer
from twisted.application import internet
from twisted.internet.error import ConnectionDone, ConnectionLost
from twisted.internet import protocol
from twisted.cred import portal
from twisted.cred import credentials
from twisted.conch.telnet import AuthenticatingTelnetProtocol
from twisted.conch.telnet import ITelnetProtocol
from twisted.conch.telnet import TelnetTransport
from twisted.conch.telnet import ECHO
from twisted.spread.pb import Avatar


class MyTelnet(Avatar):
    def __init__(self, name):
        self.name = name


@implementer(portal.IRealm)
class Realm:
    def requestAvatar(self, avatarId, mind, *interfaces):
        if ITelnetProtocol in interfaces:
            av = MyTelnet()
            av.state = "Command"
            return ITelnetProtocol, av, lambda: None
        raise NotImplementedError("Not supported by this realm")


class CanaryTelnetTransport(TelnetTransport):
    def dataReceived(self, data):
        try:
            TelnetTransport.dataReceived(self, data)
        except ValueError:
            print("Telnet client spoke weirdly, abandoning connection")
            self.loseConnection()

    def connectionLost(self, reason):
        # Avoids pointless logs on disconnect
        if reason.check(ConnectionDone) or reason.check(ConnectionLost):
            return
        TelnetTransport.connectionLost(self, reason)


class AlertAuthTelnetProtocol(AuthenticatingTelnetProtocol):
    def connectionMade(self):
        # p/Cisco telnetd/ d/router/ o/IOS/ cpe:/a:cisco:telnet/ cpe:/o:cisco:ios/a
        # NB _write() is for raw data and write() handles telnet special bytes
        self.transport._write(
            b"\xff\xfb\x01\xff\xfb\x03\xff\xfb\0\xff\xfd\0\xff\xfd\x1f\r\n"
        )
        self.transport.write(self.factory.banner)
        self.transport._write(b"User Access Verification\r\n\r\nUsername: ")

    def telnet_Password(self, line):
        # Body of this method copied from
        # twisted.conch.telnet
        username, password = self.username, line
        del self.username

        def login(ignored):
            creds = credentials.UsernamePassword(username, password)
            d = self.portal.login(creds, None, ITelnetProtocol)
            d.addCallback(self._cbLogin)
            d.addErrback(self._ebLogin)

        self.transport.wont(ECHO).addCallback(login)

        logdata = {"USERNAME": username, "PASSWORD": password}
        self.factory.canaryservice.log(logdata, transport=self.transport)
        return "Discard"


class Telnet(CanaryService):
    NAME = "telnet"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("telnet.port", default=8023))
        self.banner = config.getVal("telnet.banner", "").encode("utf8")
        self.logtype = logger.LOG_TELNET_LOGIN_ATTEMPT
        self.listen_addr = config.getVal("device.listen_addr", default="")

        if self.banner:
            self.banner += b"\n"

    def getService(self):
        r = Realm()
        p = portal.Portal(r)
        f = protocol.ServerFactory()
        f.canaryservice = self
        f.logger = self.logger
        f.banner = self.banner
        f.protocol = lambda: CanaryTelnetTransport(AlertAuthTelnetProtocol, p)
        return internet.TCPServer(self.port, f, interface=self.listen_addr)



from opencanary.modules import CanaryService
from opencanary.config import ConfigException

from twisted.protocols.policies import TimeoutMixin
from twisted.internet.protocol import Protocol
from twisted.internet.protocol import Factory
from twisted.application import internet
from random import randint

import struct
import re
import string
import random

UINT_MAX = 0xFFFFFFFF


class MySQL(Protocol, TimeoutMixin):
    HEADER_LEN = 4
    ERR_CODE_ACCESS_DENIED = 1045
    ERR_CODE_PKT_ORDER = 1156
    SQL_STATE_ACCESS_DENIED = b"28000"
    SQL_STATE_PKT_ORDER = b"08S01"

    # https://dev.mysql.com/doc/internals/en/connection-phase-packets.html#packet-Protocol::Handshake
    def __init__(self, factory):
        self._busyReceiving = False
        self._buffer = b""
        self.factory = factory
        self.threadid = factory.next_threadid()
        self.setTimeout(10)

    @staticmethod
    def build_packet(seq_id, data):
        data_len = len(data)
        if data_len > 0xFFFFFF or data_len <= 0:
            return None

        if seq_id > 0xFF or seq_id < 0:
            return None

        # chop to 3 byte int
        _length = struct.pack("<I", data_len)[:-1]
        _seq_id = struct.pack("B", seq_id)

        return _length + _seq_id + data

    @staticmethod
    def parse_auth(data):
        # https://dev.mysql.com/doc/internals/en/connection-phase-packets.html#packet-Protocol::HandshakeResponse
        offset = 4 + 4 + 1 + 23
        i = data.find(b"\x00", offset)
        if i < 0:
            return None, None

        username = data[offset:i]
        i += 1
        plen = data[i]
        i += 1
        if plen == 0:
            return username, None
        password = "".join("{:02x}".format(c) for c in data[i : i + plen])
        return username, password

    def consume_packet(self):
        if len(self._buffer) < MySQL.HEADER_LEN:
            return None, None
        length = struct.unpack("<I", self._buffer[:3] + b"\x00")[0]
        seq_id = self._buffer[3]

        # enough buffer data to consume packet?
        if len(self._buffer) < MySQL.HEADER_LEN + length:
            return seq_id, None

        payload = self._buffer[MySQL.HEADER_LEN : MySQL.HEADER_LEN + length]

        self._buffer = self._buffer[MySQL.HEADER_LEN + length :]

        return seq_id, payload

    def server_greeting(self):
        # struct.pack returns a byte string
        _threadid = struct.pack("<I", self.threadid)
        salt1, salt2 = self.gen_salt(8), self.gen_salt(12)
        data = (
            b"\x0a"
            + self.factory.canaryservice.banner
            + b"\x00"
            + _threadid
            + salt1
            + b"\x00\xff\xf7\x08\x02\x00\x0f\x80\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + salt2
            + b"\x00\x6d\x79\x73\x71\x6c\x5f\x6e\x61\x74\x69\x76\x65\x5f\x70\x61\x73\x73\x77\x6f\x72\x64\x00"
        )
        return self.build_packet(0x00, data)

    def gen_salt(self, length):
        charset = string.punctuation + string.ascii_letters + string.digits
        return b"".join(bytes(random.choice(charset), "utf-8") for _ in range(length))

    def access_denied(self, seq_id, user, password=None):
        Y = "YES" if password else "NO"
        ip = self.transport.getPeer().host
        msg = "Access denied for user '{}'@'{}' (using password: {})".format(
            user.decode("utf8"), ip, Y
        )
        return self.error_pkt(
            seq_id,
            MySQL.ERR_CODE_ACCESS_DENIED,
            MySQL.SQL_STATE_ACCESS_DENIED,
            msg.encode("utf8"),
        )

    def unordered_pkt(self, seq_id):
        msg = "Got packets out of order".encode()
        return self.error_pkt(
            seq_id, MySQL.ERR_CODE_PKT_ORDER, MySQL.SQL_STATE_PKT_ORDER, msg
        )

    def error_pkt(self, seq_id, err_code, sql_state, msg):
        data = b"\xff" + struct.pack("<H", err_code) + b"#" + sql_state + msg
        return self.build_packet(0x02, data)

    def connectionMade(self):
        self.transport.write(self.server_greeting())

    def dataReceived(self, data):
        self._buffer += data
        self.resetTimeout()

        if self._busyReceiving:
            return

        try:
            self._busyReceiving = True
            seq_id, payload = self.consume_packet()
            if seq_id is None:
                return
            elif seq_id != 1:
                # error on wrong seq_id, even if payload hasn't arrived yet
                self.transport.write(self.unordered_pkt(0x01))
                self.transport.loseConnection()
                return
            elif payload is not None:
                # seq_id == 1 and payload has arrived
                username, password = self.parse_auth(payload)
                username = bytes(str(username)[2:-1], "utf-8")
                if username:
                    logdata = {"USERNAME": username, "PASSWORD": password}
                    self.factory.canaryservice.log(logdata, transport=self.transport)
                    self.transport.write(self.access_denied(0x02, username, password))
                    self.transport.loseConnection()
        finally:
            self._busyReceiving = False

    def timeoutConnection(self):
        self.transport.abortConnection()


class SQLFactory(Factory):
    def __init__(self):
        self.threadid = randint(0, 0x0FFF)

    def next_threadid(self):
        self.threadid = (self.threadid + randint(1, 5)) & UINT_MAX
        return self.threadid

    def buildProtocol(self, addr):
        return MySQL(self)


class CanaryMySQL(CanaryService):
    NAME = "mysql"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("mysql.port", default=3306))
        self.banner = config.getVal(
            "mysql.banner", default="5.5.43-0ubuntu0.14.04.1"
        ).encode()
        self.logtype = logger.LOG_MYSQL_LOGIN_ATTEMPT
        self.listen_addr = config.getVal("device.listen_addr", default="")
        if (
            re.search("^[3456]\.[-_~.+\w]+$", self.banner.decode())  # noqa: W605
            is None
        ):
            raise ConfigException("sql.banner", "Invalid MySQL Banner")

    def getService(self):
        factory = SQLFactory()
        factory.canaryservice = self
        return internet.TCPServer(self.port, factory, interface=self.listen_addr)



from copy import deepcopy
import simplejson as json
import logging.config
import socket
import hpfeeds
import sys

from datetime import datetime
from logging.handlers import SocketHandler
from twisted.internet import reactor
import requests

from opencanary.iphelper import check_ip


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def getLogger(config):
    try:
        d = config.getVal("logger")
    except Exception:
        print("Error: config does not have 'logger' section", file=sys.stderr)
        exit(1)

    classname = d.get("class", None)
    if classname is None:
        print("Logger section is missing the class key.", file=sys.stderr)
        exit(1)

    LoggerClass = globals().get(classname, None)
    if LoggerClass is None:
        print("Logger class (%s) is not defined." % classname, file=sys.stderr)
        exit(1)

    kwargs = d.get("kwargs", None)
    if kwargs is None:
        print("Logger section is missing the kwargs key.", file=sys.stderr)
        exit(1)
    try:
        logger = LoggerClass(config, **kwargs)
    except Exception as e:
        print("An error occurred initialising the logger class", file=sys.stderr)
        print(e)
        exit(1)

    return logger


class LoggerBase(object):
    LOG_BASE_BOOT = 1000
    LOG_BASE_MSG = 1001
    LOG_BASE_DEBUG = 1002
    LOG_BASE_ERROR = 1003
    LOG_BASE_PING = 1004
    LOG_BASE_CONFIG_SAVE = 1005
    LOG_BASE_EXAMPLE = 1006
    LOG_FTP_LOGIN_ATTEMPT = 2000
    LOG_HTTP_GET = 3000
    LOG_HTTP_POST_LOGIN_ATTEMPT = 3001
    LOG_SSH_NEW_CONNECTION = 4000
    LOG_SSH_REMOTE_VERSION_SENT = 4001
    LOG_SSH_LOGIN_ATTEMPT = 4002
    LOG_SMB_FILE_OPEN = 5000
    LOG_PORT_SYN = 5001
    LOG_PORT_NMAPOS = 5002
    LOG_PORT_NMAPNULL = 5003
    LOG_PORT_NMAPXMAS = 5004
    LOG_PORT_NMAPFIN = 5005
    LOG_TELNET_LOGIN_ATTEMPT = 6001
    LOG_HTTPPROXY_LOGIN_ATTEMPT = 7001
    LOG_MYSQL_LOGIN_ATTEMPT = 8001
    LOG_MSSQL_LOGIN_SQLAUTH = 9001
    LOG_MSSQL_LOGIN_WINAUTH = 9002
    LOG_TFTP = 10001
    LOG_NTP_MONLIST = 11001
    LOG_VNC = 12001
    LOG_SNMP_CMD = 13001
    LOG_RDP = 14001
    LOG_SIP_REQUEST = 15001
    LOG_GIT_CLONE_REQUEST = 16001
    LOG_REDIS_COMMAND = 17001
    LOG_TCP_BANNER_CONNECTION_MADE = 18001
    LOG_TCP_BANNER_KEEP_ALIVE_CONNECTION_MADE = 18002
    LOG_TCP_BANNER_KEEP_ALIVE_SECRET_RECEIVED = 18003
    LOG_TCP_BANNER_KEEP_ALIVE_DATA_RECEIVED = 18004
    LOG_TCP_BANNER_DATA_RECEIVED = 18005
    LOG_USER_0 = 99000
    LOG_USER_1 = 99001
    LOG_USER_2 = 99002
    LOG_USER_3 = 99003
    LOG_USER_4 = 99004
    LOG_USER_5 = 99005
    LOG_USER_6 = 99006
    LOG_USER_7 = 99007
    LOG_USER_8 = 99008
    LOG_USER_9 = 99009

    def sanitizeLog(self, logdata):
        logdata["node_id"] = self.node_id
        logdata["local_time"] = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")
        logdata["utc_time"] = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")
        logdata["local_time_adjusted"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        if "src_host" not in logdata:
            logdata["src_host"] = ""
        if "src_port" not in logdata:
            logdata["src_port"] = -1
        if "dst_host" not in logdata:
            logdata["dst_host"] = ""
        if "dst_port" not in logdata:
            logdata["dst_port"] = -1
        if "logtype" not in logdata:
            logdata["logtype"] = self.LOG_BASE_MSG
        if "logdata" not in logdata:
            logdata["logdata"] = {}
        return logdata


class PyLogger(LoggerBase):
    """
    Generic python logging
    """

    __metaclass__ = Singleton

    def __init__(self, config, handlers, formatters={}):
        self.node_id = config.getVal("device.node_id")

        # Build config dict to initialise
        # Ensure all handlers don't drop logs based on severity level
        for h in handlers:
            handlers[h]["level"] = "NOTSET"

        logconfig = {
            "version": 1,
            "formatters": formatters,
            "handlers": handlers,
            # initialise all defined logger handlers
            "loggers": {self.node_id: {"handlers": handlers.keys()}},
        }

        try:
            logging.config.dictConfig(logconfig)
        except Exception as e:
            print("Invalid logging config", file=sys.stderr)
            print(type(e))
            print(e)
            exit(1)

        # Check if ignorelist is populated
        self.ip_ignorelist = config.getVal("ip.ignorelist", default=[])
        self.logtype_ignorelist = config.getVal("logtype.ignorelist", default=[])

        self.logger = logging.getLogger(self.node_id)

    def error(self, data):
        data["local_time"] = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S.%f")
        msg = "[ERR] %r" % json.dumps(data, sort_keys=True)
        print(msg, file=sys.stderr)
        self.logger.warn(msg)

    def log(self, logdata, retry=True):
        logdata = self.sanitizeLog(logdata)
        # Log only if not in ignorelist
        notify = True
        if "src_host" in logdata:
            for ip in self.ip_ignorelist:
                if check_ip(logdata["src_host"], ip) is True:
                    notify = False
                    break

        if "logtype" in logdata and logdata["logtype"] in self.logtype_ignorelist:
            notify = False

        if notify is True:
            self.logger.warn(json.dumps(logdata, sort_keys=True))


class SocketJSONHandler(SocketHandler):
    """Emits JSON messages over TCP delimited by newlines ('\n')"""

    def makeSocket(self, timeout=1):
        s = SocketHandler.makeSocket(self, timeout)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        return s

    def __init__(self, *args, **kwargs):
        SocketHandler.__init__(self, *args, **kwargs)
        self.retryStart = 0
        self.retryMax = 0
        self.retryFactor = 0

    def send(self, s, attempt=0):
        if attempt >= 10:
            print("Dropping log message due to too many failed sends")
            return

        if self.sock is None:
            self.createSocket()

        if self.sock:
            try:
                # TODO: Weirdly, one the other ends drops the
                # connection for the next msg, sendall still reports
                # successful write on a disconnected socket but then
                # on subsequent writes it fails correctly.
                self.sock.sendall(s.encode("utf-8"))
                return
            except socket.error:
                self.sock.close()
                self.sock = None

        # Here, we've failed to send s so retry
        reactor.callLater(1.5, lambda x: self.send(s, attempt + 1), None)

    def makePickle(self, record):
        return record.getMessage() + "\n"


class HpfeedsHandler(logging.Handler):
    def __init__(self, host, port, ident, secret, channels):
        logging.Handler.__init__(self)
        self.host = str(host)
        self.port = int(port)
        self.ident = str(ident)
        self.secret = str(secret)
        self.channels = map(str, channels)
        hpc = hpfeeds.new(self.host, self.port, self.ident, self.secret)
        hpc.subscribe(channels)
        self.hpc = hpc

    def emit(self, record):
        try:
            msg = self.format(record)
            self.hpc.publish(self.channels, msg)
        except:  # noqa: E722
            print("Error on publishing to server")


class SlackHandler(logging.Handler):
    def __init__(self, webhook_url):
        logging.Handler.__init__(self)
        self.webhook_url = webhook_url

    def generate_msg(self, alert):
        msg = {}
        msg["pretext"] = "OpenCanary Alert"
        data = json.loads(alert.msg)
        msg["fields"] = []
        for k, v in data.items():
            msg["fields"].append(
                {"title": k, "value": json.dumps(v) if type(v) is dict else v}
            )
        return {"attachments": [msg]}

    def emit(self, record):
        data = self.generate_msg(record)
        response = requests.post(self.webhook_url, json=data)
        if response.status_code != 200:
            print(
                "Error %s sending Slack message, the response was:\n%s"
                % (response.status_code, response.text)
            )


class TeamsHandler(logging.Handler):
    def __init__(self, webhook_url):
        logging.Handler.__init__(self)
        self.webhook_url = webhook_url

    def message(self, data):
        message = {
            "@type": "MessageCard",
            "@context": "http://schema.org/extensions",
            "themeColor": "49c176",
            "summary": "OpenCanary Notification",
            "title": "OpenCanary Alert",
            "sections": [{"facts": self.facts(data)}],
        }
        return message

    def facts(self, data, prefix=None):
        facts = []
        for k, v in data.items():
            key = str(k).lower() if prefix is None else prefix + "__" + str(k).lower()
            if type(v) is not dict:
                facts.append({"name": key, "value": str(v)})
            else:
                nested = self.facts(v, key)
                facts.extend(nested)
        return facts

    def emit(self, record):
        data = json.loads(record.msg)
        payload = self.message(data)
        headers = {"Content-Type": "application/json"}
        response = requests.post(self.webhook_url, headers=headers, json=payload)
        if response.status_code != 200:
            print(
                "Error %s sending Teams message, the response was:\n%s"
                % (response.status_code, response.text)
            )


def map_string(data, mapping):
    """Recursively map a python string dict to strings in a dictionary/list of strings.

    Example:
    >>> data = {'top': '%(top)s', 'nest1': {'middle': '%(middle)s', 'nest2': {'bottom': '%(bottom)s'}}}
    >>> mapping = {'top': 'one', 'middle': 'two', 'bottom': 'three'}
    >>> map_string(data, mapping)
    {'top': 'one', 'nest1': {'middle': 'two', 'nest2': {'bottom': 'three'}}}

    """
    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = map_string(value, mapping)
        return data
    if isinstance(data, (list, set, tuple)):
        return [map_string(d, mapping) for d in data]
    if isinstance(data, (str, bytes)):
        return data % mapping
    return data


class WebhookHandler(logging.Handler):
    def __init__(
        self, url, method="POST", data=None, status_code=200, ignore=None, **kwargs
    ):
        logging.Handler.__init__(self)
        self.url = url
        self.method = method
        self.data = data
        self.status_code = status_code
        self.ignore = ignore
        self.kwargs = kwargs

    def emit(self, record):
        message = self.format(record)
        if self.ignore is not None:
            if any(e in message for e in self.ignore):
                return

        mapping = {"message": message}
        if self.data is None:
            data = mapping
        else:
            if isinstance(self.data, dict):
                # Casting logging.config.ConvertingDict to a standard dict
                data = dict(self.data)
            else:
                data = self.data
            data = map_string(deepcopy(data), mapping)

        if "application/json" in self.kwargs.get("headers", {}).values():
            response = requests.request(
                method=self.method, url=self.url, json=data, **self.kwargs
            )
        else:
            response = requests.request(
                method=self.method, url=self.url, data=data, **self.kwargs
            )

        if response.status_code != self.status_code:
            print(
                "Error %s sending Requests payload, the response was:\n%s"
                % (response.status_code, response.text)
            )



from opencanary.modules import CanaryService
import twisted
import os
import time
import base64
import struct
from twisted.cred import portal, checkers, credentials
from twisted.conch import error, avatar, interfaces as conchinterfaces
from twisted.conch.ssh import factory, userauth, connection, keys, session, transport
from twisted.conch.openssh_compat import primes
from twisted.conch.ssh.common import MP
from twisted.internet import defer
from twisted.application import internet
from twisted.conch.ssh.common import NS, getNS

from zope.interface import implementer
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import dsa, rsa

SSH_PATH = "/var/tmp"

# pulled from Kippo


class HoneyPotSSHUserAuthServer(userauth.SSHUserAuthServer):
    def serviceStarted(self):
        userauth.SSHUserAuthServer.serviceStarted(self)

        us = self.transport.getHost()
        peer = self.transport.getPeer()

        logdata = {
            "LOCALVERSION": self.transport.ourVersionString,
            "REMOTEVERSION": self.transport.otherVersionString,
        }
        logtype = (
            self.transport.factory.canaryservice.logger.LOG_SSH_REMOTE_VERSION_SENT
        )
        log = self.transport.factory.canaryservice.log
        log(
            logdata,
            logtype=logtype,
            src_host=peer.address.host,
            src_port=peer.address.port,
            dst_host=us.address.host,
            dst_port=us.address.port,
        )

        self.bannerSent = False

    def sendBanner(self):
        banner = self.transport.factory.preauth_banner

        if self.bannerSent or not banner:
            return

        # data = ""
        # data = "\r\n".join(data.splitlines() + [""])
        self.transport.sendPacket(userauth.MSG_USERAUTH_BANNER, NS(banner) + NS("en"))
        self.bannerSent = True

    def auth_password(self, packet):
        """
        Password authentication.  Payload::
            string password

        Make a UsernamePassword credential and verify it with our portal.
        """
        password = getNS(packet[1:])[0]
        c = credentials.UsernamePassword(self.user, password)

        us = self.transport.getHost()
        peer = self.transport.getPeer()

        logdata = {
            "USERNAME": self.user,
            "PASSWORD": password,
            "LOCALVERSION": self.transport.ourVersionString,
            "REMOTEVERSION": self.transport.otherVersionString,
        }
        logtype = self.transport.factory.canaryservice.logger.LOG_SSH_LOGIN_ATTEMPT

        log = self.transport.factory.canaryservice.log
        log(
            logdata,
            logtype=logtype,
            src_host=peer.address.host,
            src_port=peer.address.port,
            dst_host=us.address.host,
            dst_port=us.address.port,
        )

        return self.portal.login(c, None, conchinterfaces.IConchUser).addErrback(
            self._ebPassword
        )

    def auth_publickey(self, packet):

        try:
            # extract the public key blob from the SSH packet
            key_blob = getNS(getNS(packet[1:])[1])[0]
        except:  # noqa: E722
            key_blob = "No public key found."
        try:
            # convert blob into openssh key format
            key = keys.Key.fromString(key_blob).toString("openssh")
        except:  # noqa: E722
            key = "Invalid SSH Public Key Submitted: {key_blob}".format(
                key_blob=key_blob.hex()
            )
            for keytype in [
                b"ecdsa-sha2-nistp256",
                b"ecdsa-sha2-nistp384",
                b"ecdsa-sha2-nistp521",
                b"ssh-ed25519",
            ]:
                if keytype in key_blob:
                    key = "{keytype} {keydata}".format(
                        keytype=keytype, keydata=base64.b64encode(key_blob)
                    )

            print("Key was {key}".format(key=key))

        c = credentials.SSHPrivateKey(None, None, None, None, None)

        return self.portal.login(c, None, conchinterfaces.IConchUser).addErrback(
            self._ebPassword
        )

    def ssh_USERAUTH_REQUEST(self, packet):
        self.sendBanner()
        return userauth.SSHUserAuthServer.ssh_USERAUTH_REQUEST(self, packet)


@implementer(portal.IRealm)
class HoneyPotRealm:
    def __init__(self):
        pass

    def requestAvatar(self, avatarId, mind, *interfaces):
        if conchinterfaces.IConchUser in interfaces:
            return interfaces[0], HoneyPotAvatar(avatarId, self.env), lambda: None
        else:
            raise Exception("No supported interfaces found.")


class HoneyPotTransport(transport.SSHServerTransport):

    hadVersion = False

    def connectionMade(self):
        logdata = {"SESSION": str(self.transport.sessionno)}
        logtype = self.factory.canaryservice.logger.LOG_SSH_NEW_CONNECTION
        log = self.factory.canaryservice.log
        log(logdata, transport=self.transport, logtype=logtype)

        self.interactors = []
        self.logintime = time.time()
        transport.SSHServerTransport.connectionMade(self)

    def sendKexInit(self):
        # Don't send key exchange prematurely
        if not self.gotVersion:
            return
        transport.SSHServerTransport.sendKexInit(self)

    def dataReceived(self, data):
        transport.SSHServerTransport.dataReceived(self, data)
        # later versions seem to call sendKexInit again on their own
        isLibssh = data.find(b"libssh", data.find(b"SSH-")) != -1

        if (
            (twisted.version.major < 11 or isLibssh)
            and not self.hadVersion
            and self.gotVersion
        ):
            self.sendKexInit()
            self.hadVersion = True

    def ssh_KEXINIT(self, packet):
        # print('Remote SSH version: %s' % (self.otherVersionString,))
        return transport.SSHServerTransport.ssh_KEXINIT(self, packet)

    def ssh_KEX_DH_GEX_REQUEST(self, packet):
        MSG_KEX_DH_GEX_GROUP = 31
        # We have to override this method since the original will
        # pick the client's ideal DH group size. For some SSH clients, this is
        # 8192 bits, which takes minutes to compute. Instead, we pick the minimum,
        # which on our test client was 1024.
        if self.ignoreNextPacket:
            self.ignoreNextPacket = 0
            return
        self.dhGexRequest = packet
        min, ideal, max = struct.unpack(">3L", packet)
        self.g, self.p = self.factory.getDHPrime(min)
        self.sendPacket(MSG_KEX_DH_GEX_GROUP, MP(self.p) + MP(self.g))

    # this seems to be the only reliable place of catching lost connection
    def connectionLost(self, reason):
        for i in self.interactors:
            i.sessionClosed()
        if self.transport.sessionno in self.factory.sessions:
            del self.factory.sessions[self.transport.sessionno]
        transport.SSHServerTransport.connectionLost(self, reason)

    def sendDisconnect(self, reason, desc):
        """
        Workaround for the "bad packet length" error message.

        @param reason: the reason for the disconnect.  Should be one of the
                       DISCONNECT_* values.
        @type reason: C{int}
        @param desc: a description of the reason for the disconnection.
        @type desc: C{str}
        """
        if "bad packet length" not in desc.decode():
            # With python >= 3 we can use super?
            transport.SSHServerTransport.sendDisconnect(self, reason, desc)
        else:
            self.transport.write("Protocol mismatch.\n")
            log = self.transport.factory.canaryservice.log
            log.msg("Disconnecting with error, code %s\nreason: %s" % (reason, desc))
            self.transport.loseConnection()


class HoneyPotSSHFactory(factory.SSHFactory):

    protocol = HoneyPotTransport
    services = {
        b"ssh-userauth": HoneyPotSSHUserAuthServer,
        b"ssh-connection": connection.SSHConnection,
    }

    # Special delivery to the loggers to avoid scope problems
    def logDispatch(self, sessionid, msg):
        data = {}
        data["logdata"] = msg
        self.logger.log(data)

    def __init__(self, logger=None, version=None, path=SSH_PATH, preauth_banner=None):
        # protocol^Wwhatever instances are kept here for the interact feature
        self.sessions = {}
        self.logger = logger
        self.version = version
        self.protocol.ourVersionString = version
        self.preauth_banner = preauth_banner
        rsa_pubKeyString, rsa_privKeyString = getRSAKeys(path)
        dsa_pubKeyString, dsa_privKeyString = getDSAKeys(path)
        self.publicKeys = {
            b"ssh-rsa": keys.Key.fromString(data=rsa_pubKeyString),
            b"ssh-dss": keys.Key.fromString(data=dsa_pubKeyString),
            b"rsa-sha2-512": keys.Key.fromString(data=rsa_pubKeyString),
            b"rsa-sha2-256": keys.Key.fromString(data=rsa_pubKeyString),
        }
        self.privateKeys = {
            b"ssh-rsa": keys.Key.fromString(data=rsa_privKeyString),
            b"ssh-dss": keys.Key.fromString(data=dsa_privKeyString),
            b"rsa-sha2-512": keys.Key.fromString(data=rsa_privKeyString),
            b"rsa-sha2-256": keys.Key.fromString(data=rsa_privKeyString),
        }

    def getPrimes(self):
        """
        Called when the factory is started to get Diffie-Hellman generators and
        primes to use.  Returns a dictionary mapping number of bits to lists
        of tuple of (generator, prime).

        @rtype: L{dict}
        """
        _modulis = ["/etc/ssh/moduli", "/private/etc/moduli"]
        _primes = None
        for _moduli in _modulis:
            try:
                _primes = primes.parseModuliFile(_moduli)
                break
            except IOError:
                pass
        return _primes


class HoneyPotSSHSession(session.SSHSession):
    def request_env(self, data):
        # print('request_env: %s' % (repr(data)))
        pass


@implementer(conchinterfaces.ISession)
class HoneyPotAvatar(avatar.ConchUser):
    def __init__(self, username, env):
        avatar.ConchUser.__init__(self)
        self.username = username
        self.env = env
        self.channelLookup.update({"session": HoneyPotSSHSession})

    def openShell(self, protocol):
        return

    def getPty(self, terminal, windowSize, attrs):
        return None

    def execCommand(self, protocol, cmd):
        return

    def closed(self):
        pass

    def eofReceived(self):
        pass

    def windowChanged(self, windowSize):
        self.windowSize = windowSize


def getRSAKeys(path):
    """
    Checks for existing RSA Keys, if there are none, generates a 2048 bit
    RSA key pair, saves them to a temporary location and returns the keys
    formatted as OpenSSH keys.
    """
    public_key = os.path.join(path, "id_rsa.pub")
    private_key = os.path.join(path, "id_rsa")

    if not (os.path.exists(public_key) and os.path.exists(private_key)):
        ssh_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )
        public_key_string = ssh_key.public_key().public_bytes(
            serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
        )
        private_key_string = ssh_key.private_bytes(
            serialization.Encoding.PEM,
            serialization.PrivateFormat.TraditionalOpenSSL,
            serialization.NoEncryption(),
        )
        with open(public_key, "w+b") as key_file:
            key_file.write(public_key_string)

        with open(private_key, "w+b") as key_file:
            key_file.write(private_key_string)
    else:
        with open(public_key) as key_file:
            public_key_string = key_file.read()
        with open(private_key) as key_file:
            private_key_string = key_file.read()

    return public_key_string, private_key_string


def getDSAKeys(path):
    """
    Checks for existing DSA Keys, if there are none, generates a 2048 bit
    DSA key pair, saves them to a temporary location and returns the keys
    formatted as OpenSSH keys.
    """
    public_key = os.path.join(path, "id_dsa.pub")
    private_key = os.path.join(path, "id_dsa")

    if not (os.path.exists(public_key) and os.path.exists(private_key)):
        ssh_key = dsa.generate_private_key(key_size=1024, backend=default_backend())
        public_key_string = ssh_key.public_key().public_bytes(
            serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
        )
        private_key_string = ssh_key.private_bytes(
            serialization.Encoding.PEM,
            serialization.PrivateFormat.TraditionalOpenSSL,
            serialization.NoEncryption(),
        )
        with open(public_key, "w+b") as key_file:
            key_file.write(public_key_string)
        with open(private_key, "w+b") as key_file:
            key_file.write(private_key_string)
    else:
        with open(public_key) as key_file:
            public_key_string = key_file.read()
        with open(private_key) as key_file:
            private_key_string = key_file.read()

    return public_key_string, private_key_string


@implementer(checkers.ICredentialsChecker)
class HoneypotPasswordChecker:

    credentialInterfaces = (credentials.IUsernamePassword,)

    def __init__(self, logger=None):
        self.logger = logger
        self.auth_attempt = 0

    def requestAvatarId(self, credentials):
        return defer.fail(error.UnauthorizedLogin())


@implementer(checkers.ICredentialsChecker)
class CanaryPublicKeyChecker:

    credentialInterfaces = (credentials.ISSHPrivateKey,)

    def __init__(self, logger=None):
        self.logger = logger
        self.auth_attempt = 0

    def requestAvatarId(self, credentials):
        return defer.fail(error.UnauthorizedLogin())


class CanarySSH(CanaryService):
    NAME = "ssh"

    def __init__(self, config=None, logger=None):
        CanaryService.__init__(self, config=config, logger=logger)
        self.port = int(config.getVal("ssh.port", default=22))
        self.version = config.getVal(
            "ssh.version", default="SSH-2.0-OpenSSH_5.1p1 Debian-5"
        ).encode("utf8")

        self.preauth_banner = (
            config.getVal("ssh.preauth_banner", default="")
            .replace("\\r", "\r")
            .replace("\\n", "\n")
            .encode("utf8")
        )

        # need to ensure there's a trailing CRLF
        if self.preauth_banner and not self.preauth_banner.endswith("\r\n"):
            self.preauth_banner += "\r\n"

        self.ssh_keys_path = config.getVal("ssh.key_path", default=SSH_PATH)
        self.listen_addr = config.getVal("device.listen_addr", default="")

    def getService(self):
        factory = HoneyPotSSHFactory(
            version=self.version,
            logger=self.logger,
            path=self.ssh_keys_path,
            preauth_banner=self.preauth_banner,
        )
        factory.canaryservice = self
        factory.portal = portal.Portal(HoneyPotRealm())
        factory.portal.registerChecker(HoneypotPasswordChecker(logger=factory.logger))
        factory.portal.registerChecker(CanaryPublicKeyChecker(logger=factory.logger))
        return internet.TCPServer(self.port, factory, interface=self.listen_addr)



import sys
import warnings
import os.path
from pkg_resources import resource_filename
from twisted.application import internet
from twisted.internet.protocol import Factory
from twisted.internet.protocol import DatagramProtocol

from opencanary.honeycred import buildHoneyCredHook

# Monkey-patch-replace Twisted Protocol with CanaryProtocol class
from twisted.internet import protocol


class CanaryProtocol(protocol.Protocol):
    """TCP protocols (ie. descedents of this class) gain a log method that be
    can be called with just the event data, as transport data is added here"""

    def log(self, *args, **kwargs):
        if hasattr(self, "factory") and hasattr(self.factory, "log"):
            kwargs["transport"] = self.transport
            return self.factory.log(*args, **kwargs)

        raise AttributeError(
            """Instance of %s does not have 'factory' attribute
        or factory does not have a log function."""
            % self.__class__.__name__
        )


protocol.Protocol = CanaryProtocol


class CanaryService(object):
    NAME = "baseservice"

    def __init__(self, config=None, logger=None):
        self.config = config
        self.logger = logger
        self.logtype = None

        # if config contains honeycreds, create basic honeycreds class
        self.creds = config.getVal("%s.honeycreds" % self.NAME, [])
        if self.creds:
            self.honeyCredHook = buildHoneyCredHook(self.creds)

    @classmethod
    def resource_dir(klass):
        """Read-Only module resource directory"""
        path = os.path.join("data", klass.NAME)
        return resource_filename(__name__, path)

    @classmethod
    def resource_filename(klass, *args):
        """Access read-only data (installed with package)"""
        return os.path.join(klass.resource_dir(), *args)

    def log(self, logdata, **kwargs):
        """
        Log a module event

        For brevity, protocols may pass in Twisted transport argument
        for logger to get the IPs and ports of the connection.
        """
        data = {"logtype": self.logtype, "logdata": logdata}

        logtype = kwargs.pop("logtype", None)
        if logtype:
            msg = """Passing in the logtype to log is deprecated. (In future each module will have only only logtype.)"""
            warnings.warn(msg, DeprecationWarning)
            data["logtype"] = logtype

        transport = kwargs.pop("transport", None)
        if transport:
            us = transport.getHost()
            peer = transport.getPeer()
            data["src_host"] = peer.host
            data["src_port"] = peer.port
            data["dst_host"] = us.host
            data["dst_port"] = us.port

        # otherwise the module can include IPs and ports as kwargs
        data.update(kwargs)

        # run pre-log hooks
        if getattr(self, "honeyCredHook", None):
            username = logdata.get("USERNAME", None)
            password = logdata.get("PASSWORD", None)
            if username or password:
                data["honeycred"] = self.honeyCredHook(username, password)

        self.logger.log(data)

    def getService(self):
        """Return service to be run

        This handles the easy case where the CanaryService class is
        also the Factory/Datagram class. Subclasses should override
        this if more intricracy is needed.
        """
        if isinstance(self, Factory):
            return internet.TCPServer(self.port, self)
        elif isinstance(self, DatagramProtocol):
            return internet.UDPServer(self.port, self)

        err = (
            "The class %s does not inherit from either Factory or DatagramProtocol."
            % (self.__class__.__name__)
        )
        raise Exception(err)


if sys.platform.startswith("linux"):  # noqa: C901
    from twisted.python import filepath
    from twisted.internet import inotify
    from twisted.python._inotify import INotifyError
    from twisted.internet.inotify import IN_CREATE
    import os

    class FileSystemWatcher(object):
        def __init__(self, fileName=None):
            self.path = fileName
            self.log_dir = os.path.dirname(os.path.realpath(self.path))
            self.f = None

        def reopenFiles(self, skipToEnd=True):
            if self.f:
                self.f.close()

            try:
                self.f = open(self.path)
                if skipToEnd:
                    self.f.seek(0, 2)
            except IOError:
                self.f = None

            self.notifier.startReading()
            try:
                self.notifier.ignore(filepath.FilePath(self.path))
            except KeyError:
                pass

            try:
                self.notifier.watch(
                    filepath.FilePath(self.path), callbacks=[self.onChange]
                )
            except INotifyError:
                self.notifier.watch(
                    filepath.FilePath(self.log_dir),
                    mask=IN_CREATE,
                    callbacks=[self.onDirChange],
                )

        def start(self):
            self.notifier = inotify.INotify()
            self.reopenFiles()

        def handleLines(self, lines=None):
            pass

        def processAuditLines(
            self,
        ):
            if not self.f:
                return

            lines = self.f.read().strip().split("\n")

            self.handleLines(lines=lines)

        def onChange(self, watch, path, mask):
            # print path, 'changed', mask # or do something else!
            if mask != 2:
                self.reopenFiles()

            self.processAuditLines()

        def onDirChange(self, watch, path, mask):
            # print path, ' dir changed', mask # or do something else!
            # import pdb; pdb.set_trace()
            try:
                self.notifier.ignore(filepath.FilePath(self.log_dir))
            except KeyError:
                pass

            if mask != 2:
                self.reopenFiles(skipToEnd=False)

            self.processAuditLines()
